#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

import sys
import xml.etree.ElementTree as ET
from html.parser import HTMLParser
from io import StringIO
from xml.sax.saxutils import escape, unescape

from gi.repository import Gdk, Gtk, Pango


def guess_decode(bytes, errors='strict'):
    for encoding in [sys.getfilesystemencoding(), 'utf-8', 'utf-16', 'utf-32']:
        try:
            return bytes.decode(encoding)
        except UnicodeDecodeError:
            continue
    else:
        return bytes.decode('ascii', errors=errors)


MIME = Gdk.Atom.intern('text/html', False)
# Disable serialize/deserialize registration function because it does not work
# on GTK-3, the "guint8 *data" is converted into a Gtk.TextIter
use_serialize_func = False


def _reverse_dict(dct):
    return {j: i for i, j in dct.items()}


SIZE2SCALE = {
    '1': 1 / (1.2 * 1.2 * 1.2),
    '2': 1 / (1.2 * 1.2),
    '3': 1 / 1.2,
    '4': 1,
    '5': 1.2,
    '6': 1.2 * 1.2,
    '7': 1.2 * 1.2 * 1.2,
}
SCALE2SIZE = _reverse_dict(SIZE2SCALE)
ALIGN2JUSTIFICATION = {
    'left': Gtk.Justification.LEFT,
    'center': Gtk.Justification.CENTER,
    'right': Gtk.Justification.RIGHT,
    'justify': Gtk.Justification.FILL,
}
JUSTIFICATION2ALIGN = _reverse_dict(ALIGN2JUSTIFICATION)
FAMILIES = ['normal', 'sans', 'serif', 'monospace']


def gdk_to_hex(gdk_color):
    "Convert color to 2 digit hex"
    colors = [gdk_color.red, gdk_color.green, gdk_color.blue]
    return "#" + "".join(["%02x" % (color // 256) for color in colors])


def _markup(text):
    return '<markup>%s</markup>' % text


def _strip_markup(text):
    return text[len('<markup>'):-len('</markup>')]


def _strip_newline(text):
    return ''.join(text.splitlines())


class MarkupHTMLParse(HTMLParser):
    def __init__(self):
        HTMLParser.__init__(self)
        self.root = ET.Element('markup')
        self._tags = [self.root]
        self.head = False
        self.body = True

    def handle_starttag(self, tag, attrs):
        if tag in ['b', 'i', 'u', 'div', 'font']:
            el = ET.SubElement(self._tags[-1], tag)
            for key, value in attrs:
                el.set(key, value)
            self._tags.append(el)
        elif tag == 'br':
            ET.SubElement(self._tags[-1], tag)
        elif tag == 'head':
            self.head = True

    def handle_endtag(self, tag):
        if tag in ['b', 'i', 'u', 'div', 'font']:
            el = self._tags.pop()
            assert el.tag == tag
        elif tag == 'head':
            self.head = False
        elif tag == 'body':
            self.body = False

    def handle_data(self, data):
        if self.head or not self.body:
            return
        el = self._tags[-1]
        if len(el) == 0:
            if el.text is None:
                el.text = ''
            el.text += data
        else:
            child = el[-1]
            if child.tail is None:
                child.tail = ''
            child.tail += data


def parse_markup(markup_text):
    'Return plain text and a list of start, end TextTag'
    markup_text = StringIO(_markup(markup_text))
    plain_text = ''
    tag_stack = []
    tags = []
    for event, element in ET.iterparse(markup_text, events=['start', 'end']):
        if element.tag == 'markup':
            if event == 'start' and element.text:
                plain_text += unescape(element.text)
            if event == 'end' and element.tail:
                plain_text += unescape(element.tail)
            continue
        if event == 'start':
            tag_stack.append((element, len(plain_text)))
            if element.text:
                plain_text += unescape(element.text)
        elif event == 'end':
            if element.tag == 'div':
                plain_text += '\n'
            assert tag_stack[-1][0] == element
            _, start = tag_stack.pop()
            end = len(plain_text)
            tags.append((start, end, element))
            if element.tail:
                plain_text += unescape(element.tail)
    return plain_text, tags


def normalize_markup(markup_text, method='html'):
    parser = MarkupHTMLParse()
    parser.feed(_strip_newline(markup_text))
    root = parser.root
    parent_map = {c: p for p in root.iter() for c in p}

    def order(em):
        "Re-order alphabetically tags"
        if (len(em) == 1
                and not em.text
                and em.tag not in ['div', 'markup']):
            child, = em
            if ((em.tag > child.tag or child.tag == 'div')
                    and not child.tail):
                em.tag, child.tag = child.tag, em.tag
                em.attrib, child.attrib = child.attrib, em.attrib
                if em in parent_map:
                    order(parent_map[em])

    # Add missing div for the first line
    if len(root) > 0 and root[0].tag != 'div':
        div = ET.Element('div')
        for em in list(root):
            if em.tag != 'div':
                root.remove(em)
                div.append(em)
            else:
                break
        root.insert(0, div)

    for em in root.iter():
        while em.text is None and len(em) == 1:
            dup, = em
            if dup.tag == em.tag and dup.tail is None:
                em.attrib.update(dup.attrib)
                em.remove(dup)
                em.extend(dup)
                em.text = dup.text
            else:
                break
        order(em)

    # Add missing br to empty lines
    for em in root.findall('div'):
        if em.text is None and len(em) == 0:
            em.append(ET.Element('br'))
    # TODO order attributes
    return _strip_markup(
        ET.tostring(root, encoding='utf-8', method=method).decode('utf-8'))


def find_list_delta(old, new):
    added = [e for e in new if e not in old]
    removed = [e for e in reversed(old) if e not in new]
    return added, removed


def serialize(register, content, start, end, data):
    text = ''
    if start.compare(end) == 0:
        return text
    iter_ = start
    while True:
        # Loop over line per line to get a div per line
        tags = []
        active_tags = []

        end_line = iter_.copy()
        if not end_line.ends_line():
            end_line.forward_to_line_end()
        if end_line.compare(end) > 0:
            end_line = end

        # Open div of the line
        align = None
        for tag in iter_.get_tags():
            if tag.props.justification_set:
                align = JUSTIFICATION2ALIGN[tag.props.justification]
        if align:
            text += '<div align="%s">' % align
        else:
            text += '<div>'

        while True:
            new_tags = iter_.get_tags()

            added, removed = find_list_delta(tags, new_tags)

            for tag in removed:
                if tag not in active_tags:
                    continue

                # close all open tags after this one
                while active_tags:
                    atag = active_tags.pop()
                    text += _get_html(atag, close=True)
                    if atag == tag:
                        break
                    added.insert(0, atag)

            for tag in added:
                text += _get_html(tag)
                active_tags.append(tag)

            tags = new_tags

            old_iter = iter_.copy()

            # Move to next tag toggle
            while True:
                iter_.forward_char()
                if iter_.compare(end) == 0:
                    break
                if iter_.toggles_tag():
                    break
            # Might have moved too far
            if iter_.compare(end_line) > 0:
                iter_ = end_line

            text += escape(old_iter.get_text(iter_))

            if iter_.compare(end_line) == 0:
                break

        # close any open tags
        for tag in reversed(active_tags):
            text += _get_html(tag, close=True)

        # Close the div of the line
        text += '</div>'

        iter_.forward_char()
        if iter_.compare(end) >= 0:
            break

    return normalize_markup(text)


def deserialize(register, content, iter_, text, create_tags, data):
    if not isinstance(text, str):
        text = guess_decode(text, errors='replace')
    text, tags = parse_markup(normalize_markup(text, method='xml'))
    offset = iter_.get_offset()
    content.insert(iter_, text)

    def sort_key(tag):
        start, end, element = tag
        return start, -end, element.tag

    for start, end, element in sorted(tags, key=sort_key):
        for tag in _get_tags(content, element):
            istart = content.get_iter_at_offset(start + offset)
            iend = content.get_iter_at_offset(end + offset)
            content.apply_tag(tag, istart, iend)
    return True


def setup_tags(text_buffer):
    for name, props in reversed(_TAGS):
        text_buffer.create_tag(name, **props)


_TAGS = [
    ('bold', {'weight': Pango.Weight.BOLD}),
    ('italic', {'style': Pango.Style.ITALIC}),
    ('underline', {'underline': Pango.Underline.SINGLE}),
]
_TAGS.extend([('family %s' % family, {'family': family})
              for family in FAMILIES])
_TAGS.extend([('size %s' % size, {'scale': scale})
              for size, scale in SIZE2SCALE.items()])
_TAGS.extend([('justification %s' % align, {'justification': justification})
              for align, justification in ALIGN2JUSTIFICATION.items()])


def register_foreground(text_buffer, color):
    name = 'foreground %s' % color.to_string()
    tag_table = text_buffer.get_tag_table()
    tag = tag_table.lookup(name)
    if not tag:
        tag = text_buffer.create_tag(name, foreground_rgba=color)
    return tag


def _get_tags(content, element):
    'Return tag for the element'
    tag_table = content.get_tag_table()

    if element.tag == 'b':
        yield tag_table.lookup('bold')
    elif element.tag == 'i':
        yield tag_table.lookup('italic')
    elif element.tag == 'u':
        yield tag_table.lookup('underline')
    if 'face' in element.attrib:
        tag = tag_table.lookup('family %s' % element.attrib['face'])
        if tag:
            yield tag
    size = element.attrib.get('size')
    if size in SIZE2SCALE:
        yield tag_table.lookup('size %s' % size)
    if 'color' in element.attrib:
        color = Gdk.RGBA()
        color.parse(element.attrib['color'])
        yield register_foreground(content, color)
    align = element.attrib.get('align')
    if align in ALIGN2JUSTIFICATION:
        yield tag_table.lookup('justification %s' % align)
    # TODO style background-color


def _get_html(texttag, close=False, div_only=False):
    'Return the html tag'
    # div alignment is managed at line level
    tags = []
    attrib = {}
    font = attrib['font'] = {}
    if texttag.props.weight_set and texttag.props.weight == Pango.Weight.BOLD:
        tags.append('b')
    if texttag.props.style_set and texttag.props.style == Pango.Style.ITALIC:
        tags.append('i')
    if (texttag.props.underline_set
            and texttag.props.underline == Pango.Underline.SINGLE):
        tags.append('u')
    if texttag.props.family_set and texttag.props.family:
        font['face'] = texttag.props.family
    if texttag.props.scale_set and texttag.props.scale != 1:
        font['size'] = SCALE2SIZE[texttag.props.scale]
    if (texttag.props.foreground_set
            and texttag.props.foreground_gdk != Gdk.Color(0, 0, 0)):
        font['color'] = gdk_to_hex(texttag.props.foreground_gdk)
    # TODO style background-color
    if font:
        tags.append('font')

    if close:
        return ''.join('</%s>' % t for t in tags)
    return ''.join('<%s %s>' % (t, ' '.join('%s="%s"' % (a, v)
                                            for a, v in attrib.get(
                                                t, {}).items()))
                   for t in tags)


def get_tags(content, start, end):
    'Get all tags'
    iter_ = start.copy()
    tags = set()
    while True:
        tags.update(iter_.get_tags())
        iter_.forward_char()
        if iter_.compare(end) > 0:
            iter_ = end
        if iter_.compare(end) == 0:
            break
    tags.update(iter_.get_tags())
    return tags


def remove_tags(content, start, end, *names):
    'Remove all tags starting by names'
    tags = get_tags(content, start, end)
    for tag in sorted(tags, key=lambda t: t.get_priority()):
        for name in names:
            if tag.props.name and tag.props.name.startswith(name):
                content.remove_tag(tag, start, end)


if __name__ == '__main__':
    html = '''<b>Bold</b>
 <i>Italic</i>
 <u>Underline</u>
<div><br/></div>
<div align="center">Center</div>
<div><font face="sans" size="6">Sans6<font color="#ff0000">red</font></font></div>
<div align="center"> <b> <i><u>Title</u></i> </b></div>'''  # noqa: E501

    win = Gtk.Window()
    win.set_title('HTMLTextBuffer')

    def cb(window, event):
        if use_serialize_func:
            print(text_buffer.serialize(
                text_buffer, MIME, text_buffer.get_start_iter(),
                text_buffer.get_end_iter()))
        else:
            print(serialize(
                text_buffer, text_buffer, text_buffer.get_start_iter(),
                text_buffer.get_end_iter(), None))
        Gtk.main_quit()
    win.connect('delete-event', cb)
    vbox = Gtk.VBox()
    win.add(vbox)

    text_buffer = Gtk.TextBuffer()
    text_view = Gtk.TextView()
    text_view.set_buffer(text_buffer)
    vbox.pack_start(text_view, expand=True, fill=True, padding=0)

    setup_tags(text_buffer)
    text_buffer.register_serialize_format(str(MIME), serialize, None)
    text_buffer.register_deserialize_format(str(MIME), deserialize, None)

    if use_serialize_func:
        text_buffer.deserialize(
            text_buffer, MIME, text_buffer.get_start_iter(), html)

        result = text_buffer.serialize(
            text_buffer, MIME, text_buffer.get_start_iter(),
            text_buffer.get_end_iter())
    else:
        deserialize(
            text_buffer, text_buffer, text_buffer.get_start_iter(),
            html, text_buffer.deserialize_get_can_create_tags(MIME),
            None)

        result = serialize(
            text_buffer, text_buffer, text_buffer.get_start_iter(),
            text_buffer.get_end_iter(), None)

    assert normalize_markup(html) == result, (normalize_markup(html), result)

    win.show_all()
    Gtk.main()



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import functools

from gi.repository import Gtk

import gnuhealth.common as common
from gnuhealth.action import Action
from gnuhealth.pyson import PYSONDecoder


class StateMixin(object):

    def __init__(self, *args, **kwargs):
        self.attrs = kwargs.pop('attrs')
        super(StateMixin, self).__init__(*args, **kwargs)

    def state_set(self, record):
        if record:
            state_changes = record.expr_eval(self.attrs.get('states', {}))
        else:
            state_changes = {}
        if state_changes.get('invisible', self.attrs.get('invisible')):
            self.hide()
        else:
            self.show()


class Label(StateMixin, Gtk.Label):

    def state_set(self, record):
        super(Label, self).state_set(record)
        if 'name' in self.attrs and record:
            field = record.group.fields[self.attrs['name']]
        else:
            field = None
        if not self.attrs.get('string', True):
            if field and record:
                text = field.get_client(record) or ''
            else:
                text = ''
            self.set_text(text)
        if record:
            state_changes = record.expr_eval(self.attrs.get('states', {}))
        else:
            state_changes = {}
        required = (
            (field and field.attrs.get('required'))
            or state_changes.get('required'))
        readonly = (
            (field and field.attrs.get('readonly'))
            or state_changes.get('readonly', not bool(field)))
        common.apply_label_attributes(self, readonly, required)


class VBox(StateMixin, Gtk.VBox):
    pass


class Image(StateMixin, Gtk.Image):

    def state_set(self, record):
        super(Image, self).state_set(record)
        if not record:
            return
        name = self.attrs['name']
        if name in record.group.fields:
            field = record.group.fields[name]
            name = field.get(record)
        size = int(self.attrs.get('size', 48))
        if self.attrs.get('type') == 'url':
            pixbuf = common.IconFactory.get_pixbuf_url(
                name, size=size, size_param=self.attrs.get('url_size'))
        else:
            pixbuf = common.IconFactory.get_pixbuf(name, size)
        self.set_from_pixbuf(pixbuf)


class Frame(StateMixin, Gtk.Frame):

    def __init__(self, label=None, attrs=None):
        if not label:  # label must be None to have no label widget
            label = None
        super(Frame, self).__init__(label=label, attrs=attrs)
        if not label:
            self.set_shadow_type(Gtk.ShadowType.NONE)
        self.set_border_width(0)


class ScrolledWindow(StateMixin, Gtk.ScrolledWindow):

    def state_set(self, record):
        # Force to show first to ensure it is displayed in the Notebook
        self.show()
        super(ScrolledWindow, self).state_set(record)


class Notebook(StateMixin, Gtk.Notebook):
    pass


class Expander(StateMixin, Gtk.Expander):

    def __init__(self, label=None, attrs=None):
        if not label:
            label = None
        super(Expander, self).__init__(label=label, attrs=attrs)


class Link(StateMixin, Gtk.Button):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_relief(Gtk.ReliefStyle.NONE)
        self.set_can_focus(False)
        if self.attrs.get('icon'):
            self.set_always_show_image(True)
            self.set_image(common.IconFactory.get_image(
                self.attrs['icon'], Gtk.IconSize.LARGE_TOOLBAR))
            self.set_image_position(Gtk.PositionType.TOP)
        self._current = None

    @property
    def action_id(self):
        return int(self.attrs['id'])

    def state_set(self, record):
        super().state_set(record)
        if not self.get_visible():
            return
        if record:
            if record.id < 0:
                self.hide()
                return
            data = {
                'model': record.model_name,
                'id': record.id,
                'ids': [record.id],
            }
            context = record.get_context()
            pyson_ctx = {
                'active_model': record.model_name,
                'active_id': record.id,
                'active_ids': [record.id],
            }
            self._current = record.id
        else:
            data = {}
            context = {}
            pyson_ctx = {}
            self._current = None
        pyson_ctx['context'] = context
        try:
            self.disconnect_by_func(self.__class__.clicked)
        except TypeError:
            pass
        self.connect('clicked', self.__class__.clicked, [data, context])
        action = common.RPCExecute(
            'model', 'ir.action', 'get_action_value', self.action_id,
            context=context)
        self.set_label(action['name'])

        decoder = PYSONDecoder(pyson_ctx)
        domain = decoder.decode(action['pyson_domain'])
        if action.get('pyson_search_value'):
            domain = [domain, decoder.decode(action['pyson_search_value'])]
        tab_domains = [
            (n, decoder.decode(d))
            for n, d, c in action['domains'] if c]
        if tab_domains:
            label = ('%s\n' % action['name']) + '\n'.join(
                '%s (%%s)' % n for n, _ in tab_domains)
        else:
            label = '%s (%%s)' % action['name']
        if record and self.action_id in record.links_counts:
            counter = record.links_counts[self.action_id]
            self._set_label_counter(label, counter)
        else:
            counter = [0] * (len(tab_domains) or 1)
            if record:
                record.links_counts[self.action_id] = counter
            if tab_domains:
                for i, (_, tab_domain) in enumerate(tab_domains):
                    common.RPCExecute(
                        'model', action['res_model'], 'search_count',
                        ['AND', domain, tab_domain], 0, 100, context=context,
                        callback=functools.partial(
                            self._set_count, idx=i, current=self._current,
                            counter=counter, label=label))
            else:
                common.RPCExecute(
                    'model', action['res_model'], 'search_count',
                    domain, 0, 100, context=context,
                    callback=functools.partial(
                        self._set_count, current=self._current,
                        counter=counter, label=label))

    def _set_count(self, value, idx=0, current=None, counter=None, label=''):
        if current != self._current or not self.get_parent():
            return
        try:
            count = value()
            if count > 99:
                count = '99+'
            counter[idx] = count
        except common.RPCException:
            pass
        self._set_label_counter(label, counter)

    def _set_label_counter(self, label, counter):
        self.set_label(label % tuple(counter))
        if self.attrs.get('empty') == 'hide':
            if any(counter):
                self.show()
            else:
                self.hide()

    def clicked(self, data):
        Action.execute(self.action_id, *data, keyword=True)



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from gi.repository import GObject, Gtk

from gnuhealth.common.selection import selection_shortcuts


class CellRendererCombo(Gtk.CellRendererCombo):

    def on_editing_started(self, editable, path):
        super().on_editing_started(editable, path)
        selection_shortcuts(editable)


GObject.type_register(CellRendererCombo)



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

import gettext

from gi.repository import Gtk

from gnuhealth.common import RPCException, RPCExecute, timezoned_date
from gnuhealth.common.underline import set_underline
from gnuhealth.gui.window.view_form.screen import Screen
from gnuhealth.gui.window.win_form import WinForm

_ = gettext.gettext


class Log(WinForm):

    def __init__(self, record):
        self.resource = '%s,%s' % (record.model_name, record.id)
        title = _("Logs (%s)") % record.rec_name()

        context = record.get_context()
        try:
            log, = RPCExecute(
                'model', record.model_name, 'read', [record.id],
                ['create_uid.rec_name', 'create_date',
                    'write_uid.rec_name', 'write_date'], context=context)
        except RPCException:
            return

        date_format = context.get('date_format', '%x')
        datetime_format = date_format + ' %H:%M:%S.%f'

        grid = Gtk.Grid(
            column_spacing=3, row_spacing=3, border_width=3)

        entry_model = Gtk.Entry(editable=False)
        entry_model.set_text(record.model_name)
        grid.attach(entry_model, 1, 1, 1, 1)
        label_model = Gtk.Label(
            label=set_underline(_("Model:")),
            use_underline=True, halign=Gtk.Align.END)
        label_model.set_mnemonic_widget(entry_model)
        grid.attach(label_model, 0, 1, 1, 1)

        entry_id = Gtk.Entry(editable=False)
        entry_id.set_alignment(1)
        entry_id.set_text(str(record.id))
        grid.attach(entry_id, 3, 1, 1, 1)
        label_id = Gtk.Label(
            label=set_underline(_("ID:")),
            use_underline=True, halign=Gtk.Align.END)
        label_id.set_mnemonic_widget(entry_id)
        grid.attach(label_id, 2, 1, 1, 1)

        for i, (user, user_label, date, date_label) in enumerate([
            ('create_uid.', _("Created by:"),
             'create_date', _("Created at:")),
            ('write_uid.', _("Last Modified by:"),
             'write_date', _("Last Modified at:"))], 2):
            entry_user = Gtk.Entry(editable=False, width_chars=50)
            user = log.get(user)
            if user:
                user = user.get('rec_name', '')
            entry_user.set_text(user or '')
            grid.attach(entry_user, 1, i, 1, 1)
            label_user = Gtk.Label(
                label=set_underline(user_label),
                use_underline=True, halign=Gtk.Align.END)
            label_user.set_mnemonic_widget(entry_user)
            grid.attach(label_user, 0, i, 1, 1)

            entry_date = Gtk.Entry(editable=False)
            date = log.get(date)
            if date:
                date = timezoned_date(date).strftime(datetime_format)
            entry_date.set_width_chars(len(date or ''))
            entry_date.set_text(date or '')
            grid.attach(entry_date, 3, i, 1, 1)
            label_date = Gtk.Label(
                label=set_underline(date_label),
                use_underline=True, halign=Gtk.Align.END)
            label_date.set_mnemonic_widget(entry_date)
            grid.attach(label_date, 2, i, 1, 1)

        grid.show_all()

        screen = Screen('ir.model.log', domain=[
            ('resource', '=', self.resource),
        ], mode=['tree', 'form'])
        super().__init__(screen, view_type='tree', title=title)
        screen.search_filter()

        self.win.vbox.pack_start(grid, expand=False, fill=True, padding=0)
        self.win.vbox.reorder_child(grid, 2)



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import gettext

from gi.repository import Gtk

from .widget import Widget

_ = gettext.gettext


class CheckBox(Widget):

    def __init__(self, view, attrs):
        super(CheckBox, self).__init__(view, attrs)
        self.widget = self.mnemonic_widget = Gtk.CheckButton()
        self.widget.connect('focus-out-event', lambda x, y: self._focus_out())
        self.widget.connect_after('toggled', self.sig_activate)

    def _readonly_set(self, value):
        super(CheckBox, self)._readonly_set(value)
        # TODO find a better solution to accept focus
        self.widget.set_sensitive(not value)

    def set_value(self):
        self.field.set_client(self.record, self.widget.get_active())

    def display(self):
        super(CheckBox, self).display()
        if not self.field:
            self.widget.set_active(False)
            return False
        self.widget.handler_block_by_func(self.sig_activate)
        try:
            self.widget.set_active(bool(self.field.get(self.record)))
        finally:
            self.widget.handler_unblock_by_func(self.sig_activate)



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import gettext

from gnuhealth.common import MODELACCESS
from gnuhealth.gui.window import Window

_ = gettext.gettext


def translate_view(datas):
    model = datas['model']
    Window.create(
        'ir.translation',
        res_id=False,
        domain=[('model', '=', model)],
        mode=['tree', 'form'],
        name=_('Translate view'))


def get_plugins(model):
    access = MODELACCESS['ir.translation']
    if access['read'] and access['write']:
        return [
            (_('Translate view'), translate_view),
            ]
    else:
        return []



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from gi.repository import GObject, Gtk, Pango


class CellRendererButton(Gtk.CellRenderer):
    __gproperties__ = {
        "text": (GObject.TYPE_STRING, None, "Text",
                 "Displayed text", GObject.ParamFlags.READWRITE),
        'visible': (GObject.TYPE_INT, 'Visible',
                    'Visible', 0, 10, 0, GObject.ParamFlags.READWRITE),
        'sensitive': (GObject.TYPE_INT, 'Sensitive',
                      'Sensitive', 0, 10, 0, GObject.ParamFlags.READWRITE),
    }

    __gsignals__ = {
        'clicked': (GObject.SignalFlags.RUN_LAST, GObject.TYPE_NONE,
                    (GObject.TYPE_STRING,)),
    }

    def __init__(self, text=""):
        Gtk.CellRenderer.__init__(self)
        self.text = text
        self.set_property('mode', Gtk.CellRendererMode.EDITABLE)
        self.visible = True
        self.sensitive = True

    def do_set_property(self, pspec, value):
        setattr(self, pspec.name, value)

    def do_get_property(self, pspec):
        return getattr(self, pspec.name)

    def do_render(self, cr, widget, background_area, cell_area, flags):
        if not self.visible:
            return

        if not self.sensitive:
            state = Gtk.StateFlags.INSENSITIVE
        else:
            state = Gtk.StateFlags.NORMAL

        context = widget.get_style_context()
        context.save()
        context.add_class('text-button')
        context.set_state(state)

        xpad, ypad = self.get_padding()
        x = cell_area.x + xpad
        y = cell_area.y + ypad
        w = cell_area.width - 2 * xpad
        h = cell_area.height - 2 * ypad

        Gtk.render_background(context, cr, x, y, w, h)
        Gtk.render_frame(context, cr, x, y, w, h)

        padding = context.get_padding(state)
        layout = widget.create_pango_layout(self.text)
        layout.set_width((w - padding.left - padding.right) * Pango.SCALE)
        layout.set_ellipsize(Pango.EllipsizeMode.END)
        layout.set_wrap(Pango.WrapMode.CHAR)

        lw, lh = layout.get_size()  # Can not use get_pixel_extents
        lw /= Pango.SCALE
        lh /= Pango.SCALE
        if w < lw:
            x = x + padding.left
        else:
            x = x + padding.left + 0.5 * (
                w - padding.left - padding.right - lw)
        y = y + padding.top + 0.5 * (h - padding.top - padding.bottom - lh)

        Gtk.render_layout(context, cr, x, y, layout)
        context.restore()

    def do_get_size(self, widget, cell_area=None):
        if cell_area is None:
            return (0, 0, 30, 18)
        else:
            return (cell_area.x, cell_area.y,
                    cell_area.width, cell_area.height)

    def do_start_editing(
            self, event, widget, path, background_area, cell_area, flags):
        if not self.visible or not self.sensitive:
            return
        self.emit("clicked", path)


GObject.type_register(CellRendererButton)



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import datetime
import gettext

from dateutil.parser import parse
from dateutil.relativedelta import relativedelta
from gi.repository import Gdk, GObject, Gtk

from .common import IconFactory

__all__ = [
    'Date', 'CellRendererDate', 'Time', 'CellRendererTime', 'DateTime',
    'date_parse']

_ = gettext.gettext


def _fix_format(format_):
    if '%Y' in format_:
        if (datetime.date.min.strftime('%Y') != '0001'
                and datetime.date.min.strftime('%4Y') == '0001'):
            format_ = format_.replace('%Y', '%4Y')
    return format_


def date_parse(text, format_='%x'):
    try:
        return datetime.datetime.strptime(text, format_)
    except ValueError:
        pass
    formatted_date = datetime.date(1988, 7, 16).strftime(format_)
    try:
        dayfirst = formatted_date.index('16') == 0
    except ValueError:
        dayfirst = False
    try:
        monthfirst = formatted_date.index('7') <= 1
    except ValueError:
        monthfirst = False
    yearfirst = not dayfirst and not monthfirst
    return parse(text, dayfirst=dayfirst, yearfirst=yearfirst, ignoretz=True)


class Date(Gtk.Entry):
    __gtype_name__ = 'Date'
    __gproperties__ = {
        'value': (GObject.TYPE_PYOBJECT,
                  _('Value'),
                  _('Displayed value'),
                  GObject.ParamFlags.READABLE | GObject.ParamFlags.WRITABLE),
        'format': (GObject.TYPE_STRING,
                   '%x',
                   _('Format'),
                   _('Display format'),
                   GObject.ParamFlags.READABLE | GObject.ParamFlags.WRITABLE),
    }
    __gsignals__ = {
        'date-changed': (
            GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.ACTION,
            GObject.TYPE_NONE, ()),
    }

    def __init__(self):
        self.__date = None
        self.__format = '%x'

        Gtk.Entry.__init__(self)

        self.set_width_chars(20)

        self.connect('focus-out-event', self.focus_out)
        self.connect('activate', self.activate)

        # Calendar Popup
        self.set_icon_from_pixbuf(
            Gtk.EntryIconPosition.PRIMARY,
            IconFactory.get_pixbuf('gnuhealth-date', Gtk.IconSize.MENU))
        self.set_icon_tooltip_text(
            Gtk.EntryIconPosition.PRIMARY,
            _('Open the calendar'))
        self.connect('icon-press', self.icon_press)

        self.__cal_popup = Gtk.Window(type=Gtk.WindowType.POPUP)
        self.__cal_popup.set_events(
            self.__cal_popup.get_events() | Gdk.EventMask.KEY_PRESS_MASK)
        self.__cal_popup.set_resizable(False)
        self.__cal_popup.connect('delete-event', self.cal_popup_closed)
        self.__cal_popup.connect('key-press-event', self.cal_popup_key_pressed)
        self.__cal_popup.connect('button-press-event',
                                 self.cal_popup_button_pressed)

        self.__calendar = Gtk.Calendar()
        cal_options = (
            Gtk.CalendarDisplayOptions.SHOW_DAY_NAMES
            | Gtk.CalendarDisplayOptions.SHOW_HEADING
            | Gtk.CalendarDisplayOptions.SHOW_WEEK_NUMBERS)
        self.__calendar.set_display_options(cal_options)
        self.__cal_popup.add(self.__calendar)
        self.__calendar.connect('day-selected', self.cal_popup_changed)
        self.__calendar.connect('day-selected-double-click',
                                self.cal_popup_double_click)
        self.__calendar.show()

    def parse(self):
        text = self.get_text()
        date = None
        if text:
            try:
                date = date_parse(text, self.__format).date()
            except (ValueError, OverflowError):
                pass

        self.__date = date

    def update_label(self):
        if not self.__date:
            self.set_text('')
            return
        self.set_text(self.__date.strftime(self.__format))

    def icon_press(self, entry, icon_pos, event):
        self.grab_focus()
        if icon_pos == Gtk.EntryIconPosition.PRIMARY:
            self.cal_popup_open()

    def cal_popup_open(self):
        self.parse()
        if self.__date:
            self.__calendar.select_month(
                self.__date.month - 1, self.__date.year)
            self.__calendar.select_day(self.__date.day)
        self.__cal_popup.set_transient_for(self.get_toplevel())
        # Show popup before because position needs the popup allocation
        popup_show(self.__cal_popup)
        popup_position(self, self.__cal_popup)

    def cal_popup_changed(self, calendar):
        year, month, day = self.__calendar.get_date()
        self.__date = datetime.date(year, month + 1, day)

        self.update_label()

        self.emit('date-changed')

    def cal_popup_double_click(self, calendar):
        self.cal_popup_hide()

    def cal_popup_key_pressed(self, calendar, event):
        if event.keyval != Gdk.KEY_Escape:
            return False

        self.stop_emission_by_name('key-press-event')
        self.cal_popup_hide()
        return True

    def cal_popup_button_pressed(self, calendar, event):
        child = event.window
        window = calendar.get_window()
        if child != window:
            while child:
                if child == window:
                    return False
                child = child.get_parent()
        self.cal_popup_hide()
        return True

    def cal_popup_closed(self, popup):
        self.cal_popup_hide()
        return True

    def cal_popup_hide(self):
        popup_hide(self.__cal_popup)
        self.grab_focus()
        self.emit('date-changed')

    def cal_popup_is_visible(self):
        return self.__cal_popup.is_visible()

    def focus_out(self, entry, event):
        previous_date = self.__date
        self.parse()
        self.update_label()
        if self.__date != previous_date:
            self.emit('date-changed')
        return False

    def activate(self, entry=None):
        self.parse()
        self.update_label()
        self.emit('date-changed')
        return False

    def do_set_property(self, prop, value):
        if prop.name == 'value':
            if isinstance(value, str):
                self.set_text(value)
                self.parse()
                value = self.__date
            if value:
                if isinstance(value, datetime.datetime):
                    value = value.date()
                assert isinstance(value, datetime.date), value
            self.__date = value
            self.update_label()
        elif prop.name == 'format':
            self.__format = _fix_format(value)
            self.update_label()

    def do_get_property(self, prop):
        if prop.name == 'value':
            return self.__date
        elif prop.name == 'format':
            return self.__format


GObject.type_register(Date)


class CellRendererDate(Gtk.CellRendererText):
    __gproperties__ = {
        'format': (GObject.TYPE_STRING,
                   _('Format'),
                   _('Display format'),
                   '%x',
                   GObject.ParamFlags.READABLE | GObject.ParamFlags.WRITABLE),
    }

    def __init__(self):
        self.__format = '%x'
        self.__entry = None
        self.__focus_out_id = 0

        Gtk.CellRendererText.__init__(self)

    def do_set_property(self, prop, value):
        if prop.name == 'format':
            self.__format = _fix_format(value)
            return
        Gtk.CellRendererText.set_property(self, prop, value)

    def do_get_property(self, prop):
        if prop.name == 'format':
            return self.__format
        return Gtk.CellRendererText.get_property(self, prop)

    def do_start_editing(
            self, event, widget, path, background_area, cell_area, flags):
        if not self.props.editable:
            return
        self.__entry = add_operators(Date())  # TODO add_operators has option
        self.__entry.props.format = self.props.format
        self.__entry.props.value = self.props.text
        self.__entry.set_has_frame(False)
        self.__entry.connect('editing-done', self.__editing_done)
        self.__focus_out_id = self.__entry.connect(
            'focus-out-event', self.__focus_out_event)
        # XXX focus-out-event
        self.__entry.show()
        return self.__entry

    def __editing_done(self, entry):
        self.__entry = None
        if self.__focus_out_id:
            entry.disconnect(self.__focus_out_id)
        canceled = entry.props.editing_canceled
        self.stop_editing(canceled)
        if canceled:
            return
        # TODO emit edited

    def __focus_out_event(self, entry, event):
        if entry.cal_popup_is_visible():
            return True
        entry.props.editing_canceled = True
        entry.editing_done()
        entry.remove_widget()
        return False


GObject.type_register(CellRendererDate)


class Time(Gtk.ComboBox):
    __gtype_name__ = 'Time'
    __gproperties__ = {
        'value': (GObject.TYPE_PYOBJECT,
                  _('Value'),
                  _('Displayed value'),
                  GObject.ParamFlags.READABLE | GObject.ParamFlags.WRITABLE),
        'format': (GObject.TYPE_STRING,
                   _('Format'),
                   _('Display format'),
                   '%X',
                   GObject.ParamFlags.READABLE | GObject.ParamFlags.WRITABLE),
    }
    __gsignals__ = {
        'time-changed': (
            GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.ACTION,
            GObject.TYPE_NONE, ()),
    }

    def __init__(self):
        self.__time = None
        self.__format = '%X'

        Gtk.ComboBox.__init__(self, has_entry=True)

        self.__entry = self.get_child()
        self.__entry.set_width_chars(10)

        self.__entry.connect('focus-out-event', self.focus_out)
        self.__entry.connect('activate', self.activate)
        self.connect('changed', self.changed)

        self.__model = Gtk.ListStore(
            GObject.TYPE_STRING, GObject.TYPE_PYOBJECT)
        self.update_model()
        self.set_model(self.__model)
        self.set_entry_text_column(0)

    def parse(self):
        text = self.__entry.get_text()
        time = None
        if text:
            try:
                time = date_parse(text).time()
            except (ValueError, OverflowError):
                pass

        self.__time = time

    def update_label(self):
        if self.__time is None:
            self.__entry.set_text('')
            return

        self.__entry.set_text(self.__time.strftime(self.__format))

    def update_model(self):
        self.__model.clear()
        timelist_set_list(
            self.__model, datetime.time(0, 0), datetime.time(23, 59),
            self.__format)

    def focus_out(self, entry, event):
        self.parse()
        self.update_label()
        self.emit('time-changed')
        return False

    def activate(self, entry=None):
        self.parse()
        self.update_label()
        self.emit('time-changed')
        return False

    def changed(self, combobox):
        # "changed" signal is also triggered by text editing
        # so only parse when a row is active
        if combobox.get_active_iter():
            self.parse()
            self.update_label()
            self.emit('time-changed')
        return False

    def do_set_property(self, prop, value):
        if prop.name == 'value':
            if isinstance(value, str):
                self.__entry.set_text(value)
                self.parse()
                value = self.__time
            if value:
                if isinstance(value, datetime.datetime):
                    value = value.time()
            self.__time = value
            self.update_label()
        elif prop.name == 'format':
            self.__format = _fix_format(value)
            self.update_label()
            self.update_model()

    def do_get_property(self, prop):
        if prop.name == 'value':
            return self.__time
        elif prop.name == 'format':
            return self.__format


GObject.type_register(Time)


class CellRendererTime(Gtk.CellRendererCombo):
    __gproperties__ = {
        'format': (GObject.TYPE_STRING,
                   '%X',
                   _('Format'),
                   _('Display format'),
                   GObject.ParamFlags.READABLE | GObject.ParamFlags.WRITABLE),
    }

    def __init__(self):
        self.__format = '%X'
        self.__combo = None
        self.__focus_out_id = 0

        Gtk.CellRendererText.__init__(self)

    def do_set_property(self, prop, value):
        if prop.name == 'format':
            self.__format = _fix_format(value)
            return
        Gtk.CellRendererText.set_property(self, prop, value)

    def do_get_property(self, prop):
        if prop.name == 'format':
            return self.__format
        return Gtk.CellRendererText.get_property(self, prop)

    def do_start_editing(
            self, event, widget, path, background_area, cell_area, flags):
        if not self.props.editable:
            return
        self.__combo = add_operators(Time())  # TODO add_operators has option
        self.__combo.props.format = self.props.format
        self.__combo.props.value = self.props.text
        self.__combo.props.has_frame = False
        self.__combo.connect('editing-done', self.__editing_done)
        # TODO: connect to changed
        self.__focus_out_id = self.__combo.connect(
            'focus-out-event', self.__focus_out_event)
        self.__combo.show()
        return self.__combo

    def __editing_done(self, combo):
        self.__combo = None
        canceled = combo.props.editing_canceled
        self.stop_editing(canceled)
        if canceled:
            return
        # TODO emit edited

    def __focus_out_event(self, combo, event):
        self.__editing_done(combo)
        return False


GObject.type_register(CellRendererTime)


class DateTime(Gtk.HBox):
    __gtype_name__ = 'DateTime'
    __gproperties__ = {
        'value': (GObject.TYPE_PYOBJECT,
                  _('Value'),
                  _('Displayed value'),
                  GObject.ParamFlags.READABLE | GObject.ParamFlags.WRITABLE),
        'date-format': (GObject.TYPE_STRING,
                        '%x',
                        _('Date Format'),
                        _('Displayed date format'),
                        GObject.ParamFlags.READABLE |
                        GObject.ParamFlags.WRITABLE),
        'time-format': (GObject.TYPE_STRING,
                        '%X',
                        _('Date Format'),
                        _('Displayed date format'),
                        GObject.ParamFlags.READABLE |
                        GObject.ParamFlags.WRITABLE),
    }
    __gsignals__ = {
        'datetime-changed': (
            GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.ACTION,
            GObject.TYPE_NONE, ()),
    }

    def __init__(self):
        Gtk.HBox.__init__(self, spacing=4)

        self.__date = Date()
        self.pack_start(self.__date, expand=True, fill=True, padding=0)
        self.__date.show()
        self.__date.connect('date-changed',
                            lambda e: self.emit('datetime-changed'))

        self.__time = Time()
        self.pack_start(self.__time, expand=True, fill=True, padding=0)
        self.__time.show()
        self.__time.connect('time-changed',
                            lambda e: self.emit('datetime-changed'))

    def parse(self):
        self.__date.parse()
        self.__time.parse()

    def do_set_property(self, prop, value):
        if prop.name == 'value':
            self.__date.props.value = value
            self.__time.props.value = value
        elif prop.name == 'date-format':
            self.__date.props.format = value
        elif prop.name == 'time-format':
            self.__time.props.format = value

    def do_get_property(self, prop):
        if prop.name == 'value':
            date = self.__date.props.value
            time = self.__time.props.value or datetime.time()
            if date:
                return datetime.datetime.combine(date, time)
            else:
                return
        elif prop.name == 'date-format':
            return self.__date.props.format
        elif prop.name == 'time-format':
            return self.__time.props.format

    def modify_bg(self, state, color):
        self.__date.modify_bg(state, color)
        self.__time.child.modify_bg(state, color)

    def modify_base(self, state, color):
        self.__date.modify_base(state, color)
        self.__time.child.modify_base(state, color)

    def modify_fg(self, state, color):
        self.__date.modify_fg(state, color)
        self.__time.child.modify_fg(state, color)

    def modify_text(self, state, color):
        self.__date.modify_text(state, color)
        self.__time.child.modify_text(state, color)


GObject.type_register(DateTime)


def popup_position(widget, popup):
    allocation = widget.get_allocation()
    popup_allocation = popup.get_allocation()
    x, y = widget.get_window().get_root_coords(allocation.x, allocation.y)
    display = widget.get_display()
    monitor = display.get_monitor_at_window(widget.get_window())
    monitor_geometry = monitor.get_geometry()
    if (monitor_geometry.height
            < y + allocation.height + popup_allocation.height):
        y -= popup_allocation.height
    else:
        y += allocation.height
    if monitor_geometry.width < x + popup_allocation.width:
        x -= popup_allocation.width
    popup.move(x, y)


def popup_show(popup):
    popup.show()
    popup.grab_focus()
    popup.grab_add()

    window = popup.get_window()
    display = window.get_display()
    seat = display.get_default_seat()
    seat.grab(window, Gdk.SeatCapabilities.ALL, True, None, None, None, None)


def popup_hide(popup):
    popup.hide()
    popup.grab_remove()
    window = popup.get_window()
    display = window.get_display()
    seat = display.get_default_seat()
    seat.ungrab()


def timelist_set_list(model, min_, max_, format_):
    time = min_
    delta = 30
    while time < max_:
        model.append((time.strftime(format_), time))
        hour = time.hour
        minute = time.minute + delta
        hour, minute = divmod(minute, 60)
        hour += time.hour
        if hour >= 24:
            break
        time = datetime.time(hour, minute)


def add_operators(widget):
    def key_press(editable, event):
        if not editable.get_editable():
            return False
        if event.keyval in OPERATORS:
            value = widget.props.value
            if value:
                if isinstance(value, datetime.time):
                    value = datetime.datetime.combine(
                        datetime.date.today(), value)
                try:
                    widget.props.value = value + OPERATORS[event.keyval]
                except TypeError:
                    return False
            return True
        elif event.keyval in (Gdk.KEY_KP_Equal, Gdk.KEY_equal):
            widget.props.value = datetime.datetime.now()
            return True
        return False

    if isinstance(widget, DateTime):
        for child in widget.get_children():
            add_operators(child)
        return widget
    if isinstance(widget, Gtk.ComboBox):
        editable = widget.get_child()
    else:
        editable = widget
    editable.connect('key-press-event', key_press)
    return widget


OPERATORS = {
    Gdk.KEY_S: relativedelta(seconds=-1),
    Gdk.KEY_s: relativedelta(seconds=1),
    Gdk.KEY_I: relativedelta(minutes=-1),
    Gdk.KEY_i: relativedelta(minutes=1),
    Gdk.KEY_H: relativedelta(hours=-1),
    Gdk.KEY_h: relativedelta(hours=1),
    Gdk.KEY_D: relativedelta(days=-1),
    Gdk.KEY_d: relativedelta(days=1),
    Gdk.KEY_W: relativedelta(weeks=-1),
    Gdk.KEY_w: relativedelta(weeks=1),
    Gdk.KEY_M: relativedelta(months=-1),
    Gdk.KEY_m: relativedelta(months=1),
    Gdk.KEY_Y: relativedelta(years=-1),
    Gdk.KEY_y: relativedelta(years=1),
}

if __name__ == '__main__':
    win = Gtk.Window()
    win.connect('delete-event', Gtk.main_quit)

    v = Gtk.VBox()
    v.show()

    d = add_operators(Date())
    d.show()
    v.pack_start(d, expand=False, fill=False, padding=0)

    t = add_operators(Time())
    t.show()
    v.pack_start(t, expand=False, fill=False, padding=0)

    t = add_operators(Time())
    t.props.format = '%H:%M'
    t.show()
    v.pack_start(t, expand=False, fill=False, padding=0)

    dt = add_operators(DateTime())
    dt.show()
    v.pack_start(dt, expand=False, fill=False, padding=0)

    win.add(v)
    win.show()

    Gtk.main()



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from pathlib import Path

from gi.repository import Gdk, GLib, Gtk

try:
    from gi.repository import EvinceDocument, EvinceView
    EvinceDocument.init()
except ImportError:
    EvinceDocument = EvinceView = None

from gnuhealth.common import data2pixbuf, resize_pixbuf

from .binary import BinaryMixin
from .widget import Widget


class Document(BinaryMixin, Widget):
    expand = True

    def __init__(self, view, attrs):
        super().__init__(view, attrs)

        self.widget = Gtk.VBox(spacing=3)
        self.image = Gtk.DrawingArea()
        self.widget.pack_start(
            self.image, expand=True, fill=True, padding=0)

        if EvinceView:
            self.evince_view = EvinceView.View()
            self.evince_scroll = Gtk.ScrolledWindow()
            self.evince_scroll.add(self.evince_view)
            self.widget.pack_start(
                self.evince_scroll, expand=True, fill=True, padding=0)
        else:
            self.evince_view = self.evince_scroll = None

    def _draw(self, drawing_area, cr, pixbuf):
        width = drawing_area.get_allocated_width()
        height = drawing_area.get_allocated_height()
        pixbuf = resize_pixbuf(pixbuf, width, height)
        width = (width - pixbuf.get_width()) / 2
        height = (height - pixbuf.get_height()) / 2
        Gdk.cairo_set_source_pixbuf(cr, pixbuf, width, height)
        cr.paint()

    def display(self):
        super().display()
        if self.field and self.record:
            data = self.field.get_data(self.record)
        else:
            data = None

        if not data:
            self.image.hide()
            if self.evince_view:
                self.evince_view.set_model(EvinceView.DocumentModel())
                self.evince_scroll.hide()
            return

        pixbuf = data2pixbuf(data)
        if pixbuf:
            try:
                self.image.disconnect_by_func(self._draw)
            except TypeError:
                pass
            self.image.connect('draw', self._draw, pixbuf)
            self.image.queue_draw()
            self.image.show()
            if self.evince_view:
                self.evince_view.set_model(EvinceView.DocumentModel())
                self.evince_scroll.hide()
        else:
            self.image.hide()
            if self.evince_view:
                self.evince_scroll.show()
                suffix = None
                if self.filename_field:
                    filename = self.filename_field.get(self.record)
                    if filename:
                        suffix = Path(filename).suffix
                filename = Path(self.field.get_filename(self.record, suffix))
                try:
                    document = (
                        EvinceDocument.Document.factory_get_document_full(
                            filename.as_uri(),
                            EvinceDocument.DocumentLoadFlags.NONE))
                    model = EvinceView.DocumentModel()
                    model.set_document(document)
                    self.evince_view.set_model(model)
                except GLib.GError:
                    self.evince_view.set_model(EvinceView.DocumentModel())
                    self.evince_scroll.hide()



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import gettext
import logging

from gi.repository import GLib, Gtk

from gnuhealth.common import RPCExecute
from gnuhealth.config import CONFIG
from gnuhealth.exceptions import GNUHealthError, GNUHealthServerError

_ = gettext.gettext
logger = logging.getLogger(__name__)


def get_completion(search=True, create=True):
    "Return a EntryCompletion"
    completion = Gtk.EntryCompletion()
    completion.set_match_func(lambda *a: True)
    completion.set_model(Gtk.ListStore(str, object, object))
    completion.set_text_column(0)
    completion.props.popup_set_width = False
    if search:
        completion.insert_action_markup(0, _('<i>Search...</i>'))
    if create:
        completion.insert_action_markup(1, _('<i>Create...</i>'))
    completion._allow_create = create
    return completion


def update_completion(entry, record, field, model, domain=None):
    "Update entry completion"
    def update(search_text, domain):
        if not entry.props.window:
            return False
        if search_text != entry.get_text():
            return False
        completion = entry.get_completion()
        completion_model = completion.get_model()
        if not search_text or not model:
            completion_model.clear()
            completion_model.search_text = search_text
            return False
        if getattr(completion_model, 'search_text', None) == search_text:
            return False
        if domain is None:
            domain = field.domain_get(record)
        context = field.get_search_context(record)
        order = field.get_search_order(record)

        def callback(results):
            try:
                results = results()
            except (GNUHealthError, GNUHealthServerError):
                results = []
            if search_text != entry.get_text():
                return False
            completion_model.clear()
            for result in results:
                if result['id'] is not None or completion._allow_create:
                    if result['id'] is None:
                        name = _('Create "%s"...') % result['name']
                    else:
                        name = result['name']
                    completion_model.append((
                        name, result['id'], result['defaults']))
            completion_model.search_text = search_text
            # Force display of popup
            entry.emit('changed')
        try:
            RPCExecute(
                'model', model, 'autocomplete', search_text, domain,
                CONFIG['client.limit'], order, context=context,
                process_exception=False, callback=callback)
        except Exception:
            logger.warning(
                "Unable to search for completion of %s", model,
                exc_info=True)
        return False
    search_text = entry.get_text()
    GLib.timeout_add(300, update, search_text, domain)



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import locale

from gi.repository import Gdk, GObject

from .cellrendererinteger import CellRendererInteger


class CellRendererFloat(CellRendererInteger):

    def __init__(self):
        super(CellRendererFloat, self).__init__()
        self.digits = None
        self.monetary = False
        self.convert = float

    def on_editing_started(self, editable, path):
        super().on_editing_started(editable, path)
        editable.connect('key-press-event', self.key_press_event)

    @property
    def __decimal_point(self):
        return locale.localeconv()['decimal_point']

    @property
    def __thousands_sep(self):
        return locale.localeconv()['thousands_sep']

    def key_press_event(self, widget, event):
        for name in ('KP_Decimal', 'KP_Separator'):
            if event.keyval == Gdk.keyval_from_name(name):
                text = self.__decimal_point
                try:
                    start_pos, end_pos = widget.get_selection_bounds()
                except ValueError:
                    start_pos = widget.get_position()
                    end_pos = None
                if self._can_insert_text(widget, text, start_pos, end_pos):
                    buffer_ = widget.get_buffer()
                    if end_pos:
                        buffer_.delete_text(start_pos, end_pos - start_pos)
                    buffer_.insert_text(start_pos, text, len(text))
                    widget.set_position(widget.get_position() + len(text))
                return True

    def _can_insert_text(self, entry, new_text, start_pos, end_pos=None):
        value = entry.get_text()
        if end_pos is None:
            end_pos = start_pos
        new_value = value[:start_pos] + new_text + value[end_pos:]
        if new_value not in {'-', self.__decimal_point, self.__thousands_sep}:
            try:
                value = self.convert(
                    locale.delocalize(new_value, self.monetary))
            except ValueError:
                return False
            if (value and self.digits is not None
                    and round(value, self.digits[1]) != value):
                return False
        return True


GObject.type_register(CellRendererFloat)



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

import gettext
import json
import logging
import os
import sys
import threading
import traceback
import webbrowser
from functools import partial
from urllib.parse import parse_qsl, unquote, urlparse

from gi.repository import Gdk, GdkPixbuf, Gio, GLib, Gtk

import gnuhealth.common as common
import gnuhealth.plugins
import gnuhealth.rpc as rpc
import gnuhealth.translate as translate
from gnuhealth.action import Action
from gnuhealth.common import RPCContextReload, RPCException, RPCExecute
from gnuhealth.common.cellrendererclickablepixbuf import (
    CellRendererClickablePixbuf)
from gnuhealth.config import CONFIG, GNUHEALTH_ICON, get_config_dir
from gnuhealth.exceptions import GNUHealthError, GNUHealthServerUnavailable
from gnuhealth.gui.window import Window
from gnuhealth.jsonrpc import object_hook
from gnuhealth.pyson import PYSONDecoder

# GNUHealth block: client version and other related imports
import time
import platform
from gnuhealth.gui.window.activity import Activity
from gnuhealth import __version__
import ast

_ = gettext.gettext
logger = logging.getLogger(__name__)
_PRIORITIES = [getattr(Gio.NotificationPriority, p)
               for p in ('LOW', 'NORMAL', 'HIGH', 'URGENT')]


class Main(Gtk.Application):
    window = None
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance

    def do_startup(self):
        Gtk.Application.do_startup(self)

        action = Gio.SimpleAction.new('preferences', None)
        action.connect('activate', lambda *a: self.preferences())
        self.add_action(action)
        self.set_accels_for_action('app.preferences', ['<Primary>comma'])

        action = Gio.SimpleAction.new('menu-search', None)
        action.connect(
            'activate', lambda *a: self.global_search_entry.grab_focus())
        self.add_action(action)
        self.set_accels_for_action('app.menu-search', ['<Primary>k'])

        action = Gio.SimpleAction.new('menu-toggle', None)
        action.connect('activate', lambda *a: self.menu_toggle())
        self.add_action(action)
        self.set_accels_for_action('app.menu-toggle', ['<Primary>m'])

        toolbar_variant = GLib.Variant.new_string(
            CONFIG['client.toolbar'] or 'both')
        action = Gio.SimpleAction.new_stateful(
            'toolbar', toolbar_variant.get_type(), toolbar_variant)
        action.connect('change-state', self.on_change_toolbar)
        self.add_action(action)

        def on_change_action_boolean(action, value, key):
            action.set_state(value)
            CONFIG[key] = value.get_boolean()

        for name, key in [
                ('save-tree-width', 'client.save_tree_width'),
                ('save-tree-state', 'client.save_tree_state'),
                ('code-scanner-sound', 'client.code_scanner_sound'),
                ('spell-checking', 'client.spellcheck'),
        ]:
            variant = GLib.Variant.new_boolean(CONFIG[key])
            action = Gio.SimpleAction.new_stateful(name, None, variant)
            action.connect('change-state', on_change_action_boolean, key)
            self.add_action(action)

        action = Gio.SimpleAction.new('tab-previous', None)
        action.connect('activate', lambda *a: self.win_prev())
        self.add_action(action)
        self.set_accels_for_action('app.tab-previous', ['<Primary><Shift>Tab'])

        action = Gio.SimpleAction.new('tab-next', None)
        action.connect('activate', lambda *a: self.win_next())
        self.add_action(action)
        self.set_accels_for_action('app.tab-next', ['<Primary>Tab'])

        action = Gio.SimpleAction.new('search-limit', None)
        action.connect('activate', lambda *a: self.edit_limit())
        self.add_action(action)

        action = Gio.SimpleAction.new('documentation', None)
        action.connect('activate', lambda *a: common.open_documentation())
        self.add_action(action)

        action = Gio.SimpleAction.new('donation', None)
        action.connect('activate', lambda *a: common.open_donation())
        self.add_action(action)

        self._shortcuts = None
        action = Gio.SimpleAction.new('shortcuts', None)
        action.connect('activate', lambda *a: self.shortcuts())
        self.add_action(action)
        self.set_accels_for_action('app.shortcuts', ['<Primary>F1'])

        action = Gio.SimpleAction.new('about', None)
        action.connect('activate', lambda *a: self.about())
        self.add_action(action)

        action = Gio.SimpleAction.new('quit', None)
        action.connect('activate', self.on_quit)
        self.add_action(action)
        self.set_accels_for_action('app.quit', ['<Primary>q'])

        # GNU Health block
        # Add Activity Window action
        action = Gio.SimpleAction.new('activity', None)
        action.connect('activate', lambda *a: self.activity_toggle())
        self.add_action(action)
        self.set_accels_for_action('app.activity', ['<Shift>F12'])

        # GNU Health Command Line focus
        action = Gio.SimpleAction.new('gh_cli', None)
        action.connect('activate', lambda *a: self.gnuhealth_cli())
        self.add_action(action)
        self.set_accels_for_action('app.gh_cli', ['<Primary><Shift>Z'])

        # GNU Health Block: Set Main Menu menubar
        main_menu = Gio.Menu.new()
        file_menu = Gio.Menu.new()
        settings_menu = Gio.Menu.new()
        help_menu = Gio.Menu.new()
        toolbar = Gio.Menu.new()
        main_menu.append_submenu(_("File"), file_menu)
        main_menu.append_submenu(_("Settings"), settings_menu)
        main_menu.append_submenu(_("Help"), help_menu)

        settings_menu.append(_("Preferences..."), 'app.preferences')

        settings_menu.append_submenu(_("Toolbar"), toolbar)
        toolbar.append(_("Default"), 'app.toolbar::default')
        toolbar.append(_("Text and Icons"), 'app.toolbar::both')
        toolbar.append(_("Text"), 'app.toolbar::text')
        toolbar.append(_("Icons"), 'app.toolbar::icons')

        form = Gio.Menu.new()
        settings_menu.append_submenu(_("Form"), form)
        form.append(_("Save Width/Height"), 'app.save-width-height')
        form.append(_("Save Tree State"), 'app.save-tree-state')
        form.append(_("Spell Checking"), 'app.spell-checking')
        form.append(_("Play Sound for Code Scanner"), 'app.code-scanner-sound')

        settings_menu.append(_("Search Limit..."), 'app.search-limit')
        settings_menu.append(_("Email..."), 'app.email')

        help_menu.append(_("Documentation..."), 'app.documentation')
        help_menu.append(_("Keyboard Shortcuts..."), 'app.shortcuts')
        help_menu.append(_("About..."), 'app.about')
        help_menu.append(
            _("Donate to GNU Health") + " " + "\N{heavy black heart}",
            'app.donation')

        file_menu.append(_("Quit"), 'app.quit')

        self.set_menubar(main_menu)

    def do_activate(self):
        if self.window:
            self.window.present()
            return

        self.window = Gtk.ApplicationWindow(application=self)
        title = self.get_main_title()
        self.window.set_title(title)
        self.window.set_default_size(1280, 960)
        self.window.maximize()
        self.window.set_position(Gtk.WindowPosition.CENTER)
        self.window.set_resizable(True)
        self.window.set_icon(GNUHEALTH_ICON)
        self.window.connect("destroy", self.on_quit)
        self.window.connect("delete_event", self.on_quit)
        common.setup_window(self.window)

        # GNU Health
        # Disable title bar
        # self.window.set_titlebar(self.header)
        self.header = Gtk.HeaderBar.new()
        self.header.set_show_close_button(True)
        self.set_title()

        menu = Gtk.Button.new()
        menu.set_relief(Gtk.ReliefStyle.NONE)
        menu.set_image(
            common.IconFactory.get_image(
                'gnuhealth-menu', Gtk.IconSize.BUTTON))
        menu.connect('clicked', self.menu_toggle)
        self.header.pack_start(menu)

        favorite = Gtk.MenuButton.new()
        favorite.set_relief(Gtk.ReliefStyle.NONE)
        favorite.set_image(common.IconFactory.get_image(
            'gnuhealth-bookmarks', Gtk.IconSize.BUTTON))
        self.menu_favorite = Gtk.Menu.new()
        favorite.set_popup(self.menu_favorite)
        favorite.connect('button-press-event', self.favorite_set)
        # self.header.pack_start(favorite)

        # self.set_global_search()
        # self.header.pack_start(self.global_search_entry)

        self.accel_group = Gtk.AccelGroup()
        self.window.add_accel_group(self.accel_group)

        Gtk.AccelMap.add_entry(
            '<gnuhealth>/Form/New', Gdk.KEY_N, Gdk.ModifierType.CONTROL_MASK)
        Gtk.AccelMap.add_entry(
            '<gnuhealth>/Form/Save', Gdk.KEY_S, Gdk.ModifierType.CONTROL_MASK)
        Gtk.AccelMap.add_entry(
            '<gnuhealth>/Form/Duplicate', Gdk.KEY_D,
            Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK)
        Gtk.AccelMap.add_entry(
            '<gnuhealth>/Form/Delete',
            Gdk.KEY_D, Gdk.ModifierType.CONTROL_MASK)
        Gtk.AccelMap.add_entry(
            '<gnuhealth>/Form/Next',
            Gdk.KEY_Down, Gdk.ModifierType.CONTROL_MASK)
        Gtk.AccelMap.add_entry(
            '<gnuhealth>/Form/Previous',
            Gdk.KEY_Up, Gdk.ModifierType.CONTROL_MASK)
        Gtk.AccelMap.add_entry(
            '<gnuhealth>/Form/Switch View', Gdk.KEY_L,
            Gdk.ModifierType.CONTROL_MASK)
        Gtk.AccelMap.add_entry(
            '<gnuhealth>/Form/Close', Gdk.KEY_W, Gdk.ModifierType.CONTROL_MASK)
        Gtk.AccelMap.add_entry(
            '<gnuhealth>/Form/Reload',
            Gdk.KEY_R, Gdk.ModifierType.CONTROL_MASK)
        Gtk.AccelMap.add_entry(
            '<gnuhealth>/Form/Attachments', Gdk.KEY_T,
            Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK)
        Gtk.AccelMap.add_entry(
            '<gnuhealth>/Form/Notes', Gdk.KEY_O,
            Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK)
        Gtk.AccelMap.add_entry(
            '<gnuhealth>/Form/Relate', Gdk.KEY_R,
            Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK)
        Gtk.AccelMap.add_entry(
            '<gnuhealth>/Form/Actions',
            Gdk.KEY_E, Gdk.ModifierType.CONTROL_MASK)
        Gtk.AccelMap.add_entry(
            '<gnuhealth>/Form/Report',
            Gdk.KEY_P, Gdk.ModifierType.CONTROL_MASK)
        Gtk.AccelMap.add_entry(
            '<gnuhealth>/Form/Email', Gdk.KEY_E,
            Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK)
        Gtk.AccelMap.add_entry(
            '<gnuhealth>/Form/Search', Gdk.KEY_F,
            Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK)

        Gtk.AccelMap.load(os.path.join(get_config_dir(), 'accel.map'))

        self.tooltips = common.Tooltips()

        self.vbox = Gtk.VBox()
        self.window.add(self.vbox)

        self.buttons = {}

        self.info = Gtk.VBox()
        self.vbox.pack_start(self.info, expand=False, fill=True, padding=0)
        self.pane = Gtk.HPaned()
        self.vbox.pack_start(self.pane, expand=True, fill=True, padding=0)
        self.pane.set_position(int(CONFIG['menu.pane']))

        self.menu_screen = None
        self.menu = Gtk.VBox()
        self.menu.set_vexpand(True)
        self.pane.add1(self.menu)

        self.notebook = Gtk.Notebook()
        self.notebook.popup_enable()
        self.notebook.set_scrollable(True)
        self.notebook.connect_after('switch-page', self._sig_page_changt)
        self.pane.add2(self.notebook)

        self.window.show_all()

        self.pages = []
        self.previous_pages = {}
        self.current_page = 0
        self.last_page = 0
        self.dialogs = []

        # Register plugins
        gnuhealth.plugins.register()

        # Initialize GNU Health environment
        # GNUHEALTH CLI
        self.cli = self.statusbar = self.footer_contents = self.footer = None
        self.cli_position = CONFIG['client.cli_position']

        self.init_gnuhealth_env()

        self.set_title()  # Adds username/profile while password is asked
        try:
            common.get_credentials()
        except Exception as exception:
            if not isinstance(exception, GNUHealthError):
                common.error(exception, traceback.format_exc())
            elif exception.faultCode == 'SessionFailed':
                common.message(
                    _("Could not get a session."),
                    msg_type=Gtk.MessageType.ERROR)
            return self.quit()
        self.get_preferences()

        # GNUHealth Block
        # Add connection successful entry to the Activity log
        msg = "Connected to GNU Health Server"
        self.activity_log_entry(msg, 'info')

    def gnuhealth_cli(self):
        self.cli.grab_focus()

    def do_command_line(self, cmd):
        self.do_activate()
        arguments = cmd.get_arguments()
        if len(arguments) > 1:
            url = arguments[1]
            self.open_url(url)
        return True

    def do_shutdown(self):
        Gtk.Application.do_shutdown(self)
        common.Logout()
        CONFIG.save()
        Gtk.AccelMap.save(os.path.join(get_config_dir(), 'accel.map'))

    def on_quit(self, *args):
        try:
            if not self.close_pages():
                return True
        except GNUHealthServerUnavailable:
            pass
        rpc.logout()
        self.quit()

    def add_window(self, window):
        super().add_window(window)
        common.setup_window(window)

    def set_global_search(self):
        self.global_search_entry = Gtk.Entry.new()
        self.global_search_entry.set_width_chars(20)
        global_search_completion = Gtk.EntryCompletion()
        global_search_completion.set_match_func(lambda *a: True)
        global_search_completion.set_model(
            Gtk.ListStore(GdkPixbuf.Pixbuf, str, str, int, str))
        pixbuf_cell = Gtk.CellRendererPixbuf()
        global_search_completion.pack_start(pixbuf_cell, expand=True)
        global_search_completion.add_attribute(pixbuf_cell, 'pixbuf', 0)
        text_cell = Gtk.CellRendererText()
        global_search_completion.pack_start(text_cell, expand=True)
        global_search_completion.add_attribute(text_cell, "markup", 1)
        global_search_completion.props.popup_set_width = True

        self.global_search_entry.set_completion(global_search_completion)

        def match_selected(completion, model, iter_):
            model, record_id, model_name = model.get(iter_, 2, 3, 4)
            if model == self.menu_screen.model_name:
                # ids is not defined to prevent to add suffix
                Action.exec_keyword('tree_open', {
                    'model': model,
                    'id': record_id,
                })
            else:
                Window.create(
                    model,
                    res_id=record_id,
                    mode=['form', 'tree'],
                    name=model_name)
            self.global_search_entry.set_text('')
            return True

        global_search_completion.connect('match-selected', match_selected)

        def update(widget, search_text, callback=None):
            def end():
                if callback:
                    callback()
                return False
            if search_text != widget.get_text():
                return end()
            gmodel = global_search_completion.get_model()
            if not search_text or not gmodel:
                gmodel.clear()
                gmodel.search_text = search_text
                return end()
            if getattr(gmodel, 'search_text', None) == search_text:
                return end()

            def set_result(result):
                try:
                    result = result()
                except RPCException:
                    result = []
                if search_text != widget.get_text():
                    if callback:
                        callback()
                    return False
                gmodel.clear()
                for r in result:
                    _, model, model_name, record_id, record_name, icon = r
                    if icon:
                        text = common.to_xml(record_name)
                        pixbuf = common.IconFactory.get_pixbuf(
                            icon, Gtk.IconSize.BUTTON)
                    else:
                        text = '<b>%s:</b>\n %s' % (
                            common.to_xml(model_name),
                            common.to_xml(record_name))
                        pixbuf = None
                    gmodel.append([pixbuf, text, model, record_id, model_name])
                gmodel.search_text = search_text
                # Force display of popup
                widget.emit('changed')
                end()

            RPCExecute(
                'model', 'ir.model', 'global_search', search_text,
                CONFIG['client.limit'], self.menu_screen.model_name,
                context=self.menu_screen.context, callback=set_result)
            return False

        def changed(widget):
            search_text = widget.get_text()
            GLib.timeout_add(300, update, widget, search_text)

        def activate(widget):
            def message():
                gmodel = global_search_completion.get_model()
                if not len(gmodel):
                    common.message(_('No result found.'))
                else:
                    widget.emit('changed')
            search_text = widget.get_text()
            update(widget, search_text, message)

        self.global_search_entry.connect('changed', changed)
        self.global_search_entry.connect('activate', activate)

    def set_title(self, value=''):
        if CONFIG['login.profile']:
            login_info = CONFIG['login.profile']
        else:
            login_info = '%s@%s/%s' % (
                CONFIG['login.login'],
                CONFIG['login.host'],
                CONFIG['login.db'])
        titles = []
        if value:
            titles.append(value)
        titles.append(CONFIG['client.title'])
        self.header.set_title(' - '.join(titles))
        self.header.set_subtitle(login_info)
        try:
            style_context = self.header.get_style_context()
        except AttributeError:
            pass
        else:
            for name in style_context.list_classes():
                if name.startswith('profile-'):
                    style_context.remove_class(name)
            if CONFIG['login.profile']:
                style_context.add_class(
                    'profile-%s' % CONFIG['login.profile'])

    def favorite_set(self, *args):
        if self.menu_favorite.get_children():
            return

        def _action_favorite(widget, id_):
            event = Gtk.get_current_event()
            allow_similar = (
                event.state & Gdk.ModifierType.MOD1_MASK
                or event.state & Gdk.ModifierType.SHIFT_MASK)
            with Window(allow_similar=allow_similar):
                # ids is not defined to prevent to add suffix
                Action.exec_keyword('tree_open', {
                    'model': self.menu_screen.model_name,
                    'id': id_,
                })

        def _manage_favorites(widget):
            Window.create(
                self.menu_screen.model_name + '.favorite',
                mode=['tree', 'form'],
                name=_("Favorites"))
        try:
            favorites = RPCExecute(
                'model',
                self.menu_screen.model_name + '.favorite', 'get',
                process_exception=False)
        except Exception:
            return False
        for id_, name, icon in favorites:
            menuitem = Gtk.MenuItem(label=name)
            menuitem.connect('activate', _action_favorite, id_)
            self.menu_favorite.add(menuitem)
        self.menu_favorite.add(Gtk.SeparatorMenuItem())
        manage_favorites = Gtk.MenuItem(
            label=_("Manage..."),
            use_underline=True)
        manage_favorites.connect('activate', _manage_favorites)
        self.menu_favorite.add(manage_favorites)
        self.menu_favorite.show_all()

    def favorite_unset(self):
        for child in self.menu_favorite.get_children():
            self.menu_favorite.remove(child)

    def on_change_toolbar(self, action, value):
        action.set_state(value)
        option = value.get_string()
        CONFIG['client.toolbar'] = option
        if option == 'default':
            barstyle = False
        elif option == 'both':
            barstyle = Gtk.ToolbarStyle.BOTH
        elif option == 'text':
            barstyle = Gtk.ToolbarStyle.TEXT
        elif option == 'icons':
            barstyle = Gtk.ToolbarStyle.ICONS
        for page_idx in range(self.notebook.get_n_pages()):
            page = self.get_page(page_idx)
            if page.toolbar:
                page.toolbar.set_style(barstyle)

    def edit_limit(self):
        from gnuhealth.gui.window.limit import Limit
        Limit().run()

    def win_next(self):
        page = self.notebook.get_current_page()
        if page == len(self.pages) - 1:
            page = -1
        self.notebook.set_current_page(page + 1)

    def win_prev(self):
        page = self.notebook.get_current_page()
        self.notebook.set_current_page(page - 1)

    def get_preferences(self):
        RPCContextReload()
        try:
            prefs = RPCExecute('model', 'res.user', 'get_preferences', False)
        except RPCException:
            prefs = {}

        targets = [
            common.IconFactory.load_icons,
            common.MODELACCESS.load_models,
            common.MODELHISTORY.load_history,
            common.MODELNOTIFICATION.load_names,
            common.VIEW_SEARCH.load_searches,
        ]
        if CONFIG['thread']:
            threads = []
            for target in targets:
                t = threading.Thread(target=target)
                threads.append(t)
                t.start()
            for t in threads:
                t.join()
        else:
            for target in targets:
                target()

        if prefs and 'language_direction' in prefs:
            translate.set_language_direction(prefs['language_direction'])
            CONFIG['client.language_direction'] = \
                prefs['language_direction']
        self.sig_win_menu(prefs=prefs)
        for action_id in prefs.get('actions', []):
            Action.execute(action_id, {})
        self.set_title(prefs.get('status_bar', ''))
        if prefs and 'language' in prefs:
            translate.setlang(prefs['language'], prefs.get('locale'))
            CONFIG['client.lang'] = prefs['language']
        common.MODELNAME.clear()
        # Set placeholder after language is set to get correct translation
        self.global_search_entry.set_placeholder_text(_("Global Search"))
        CONFIG.save()

    def preferences(self):
        from gnuhealth.gui.window.preference import Preference
        if not self.close_pages():
            return False
        Preference(rpc._USER, self.get_preferences)

    def sig_win_close(self, widget=None):
        self._sig_remove_book(
            widget,
            self.notebook.get_nth_page(self.notebook.get_current_page()))

    def close_pages(self):
        if self.notebook.get_n_pages():
            if not common.sur(
                    _('The following action requires to close all tabs.\n'
                      'Do you want to continue?')):
                return False
        res = True
        while res:
            wid = self.get_page()
            if wid:
                if not wid.sig_close():
                    return False
                res = self._win_del()
            else:
                res = False
        if self.menu_screen:
            self.menu_screen.save_tree_state()
        if self.menu.get_visible():
            CONFIG['menu.pane'] = self.pane.get_position()
        return True

    def about(self):
        from gnuhealth.gui.window.about import About
        About()

    def shortcuts(self):
        if self._shortcuts:
            self._shortcuts.destroy()
        self._shortcuts = window = Gtk.ShortcutsWindow()
        window.set_transient_for(common.get_toplevel_window())
        window.set_position(Gtk.WindowPosition.CENTER_ALWAYS)
        window.set_destroy_with_parent(True)
        self.add_window(window)

        section = Gtk.ShortcutsSection()
        section.props.section_name = 'app'
        section.props.title = _("Application Shortcuts")
        section.props.visible = True

        group = Gtk.ShortcutsGroup()
        group.props.title = _("Global")
        section.add(group)

        for action, title in [
                ('app.preferences', _("Preferences")),
                ('app.menu-search', _("Search menu")),
                ('app.menu-toggle', _("Toggle menu")),
                ('app.tab-previous', _("Previous tab"), ),
                ('app.tab-next', _("Next tab")),
                ('app.shortcuts', _("Shortcuts")),
                ('app.quit', _("Quit")),
        ]:
            shortcut = Gtk.ShortcutsShortcut()
            shortcut.props.title = title
            accels = self.get_accels_for_action(action)
            if accels:
                shortcut.props.accelerator = ' '.join(accels)
                group.add(shortcut)

        window.add(section)

        for name, title, groups in [
                ('entry', _("Edition Shortcuts"), [
                    (_("Text Entries"), [
                        ('<Primary>x', _("Cut selected text")),
                        ('<Primary>c', _("Copy selected text")),
                        ('<Primary>v', _("Paste copied text")),
                        ('Tab', _("Next entry")),
                        ('<Shift>Tab', _("Previous entry")),
                    ]),
                    (_("Relation Entries"), [
                        ('F3', _("Create new relation")),
                        ('F2', _("Open/Search relation")),
                    ]),
                    (_("List Entries"), [
                        ('F4', _("Switch view")),
                        ('F3', _("Create/Select new line")),
                        ('F2', _("Open relation")),
                        ('Delete',
                         _("Mark line for deletion/removal")),
                        ('<Ctrl>Delete', _("Mark line for removal")),
                        ('Insert', _("Unmark line for deletion")),
                    ]),
                ]),
                ('tree', _("List/Tree Shortcuts"), [
                    (_("Move Cursor"), [
                        ('Right', _("Move right")),
                        ('Left', _("Move left")),
                        ('Up', _("Move up")),
                        ('Down', _("Move down")),
                        ('Page_Up', _("Move up of one page")),
                        ('Page_Down', _("Move down of one page")),
                        ('Home', _("Move to top")),
                        ('End', _("Move to bottom")),
                        ('BackSpace', _("Move to parent")),
                    ]),
                    (_("Edition"), [
                        ('<Primary>c', _("Copy selected rows")),
                        ('<Primary>v', _("Insert copied rows")),
                    ]),
                    (_("Selection"), [
                        ('<Ctrl>a <Ctrl>slash', _("Select all")),
                        ('<Shift><Ctrl>a <Shift><Ctrl>slash',
                         _("Unselect all")),
                        ('BackSpace', _("Select parent")),
                        ('space', _("Select/Activate current row")),
                        ('<Shift>space Return',
                         _("Select/Activate current row")),
                        ('<Ctrl>space', _("Toggle selection")),
                    ]),
                    (_("Expand/Collapse"), [
                        ('plus', _("Expand row")),
                        ('minus', _("Collapse row")),
                        ('space', _("Toggle row")),
                        ('<Shift>Left', _("Collapse all rows")),
                        ('<Shift>Right', _("Expand all rows")),
                    ]),
                ]),
        ]:
            section = Gtk.ShortcutsSection()
            section.props.section_name = name
            section.props.title = title
            section.props.visible = True

            for title, shortcuts in groups:
                group = Gtk.ShortcutsGroup()
                group.props.title = title
                section.add(group)

                for accelerator, title in shortcuts:
                    shortcut = Gtk.ShortcutsShortcut()
                    shortcut.props.title = title
                    shortcut.props.accelerator = accelerator
                    group.add(shortcut)

            window.add(section)

        window.show_all()

    def menu_toggle(self, *args):
        if self.menu.get_visible():
            CONFIG['menu.pane'] = self.pane.get_position()
            self.pane.set_position(0)
            self.notebook.grab_focus()
            self.menu.set_visible(False)
        else:
            self.pane.set_position(int(CONFIG['menu.pane']))
            self.menu.set_visible(True)
            if self.menu_screen:
                self.menu_screen.set_cursor()

    def menu_row_activate(self):
        screen = self.menu_screen
        record_id = (
            screen.current_record.id
            if screen.current_record else None)
        # ids is not defined to prevent to add suffix
        return Action.exec_keyword('tree_open', {
            'model': screen.model_name,
            'id': record_id,
        }, warning=False)

    def sig_win_menu(self, prefs=None):
        from gnuhealth.gui.window.view_form.screen import Screen

        if not prefs:
            try:
                prefs = RPCExecute('model', 'res.user', 'get_preferences',
                                   False)
            except RPCException:
                return False

        if self.menu_screen:
            self.menu_screen.save_tree_state()
        for child in self.menu.get_children():
            self.menu.remove(child)

        action = PYSONDecoder().decode(prefs['pyson_menu'])
        view_ids = []
        if action.get('views', []):
            view_ids = [x[0] for x in action['views']]
        elif action.get('view_id', False):
            view_ids = [action['view_id'][0]]
        ctx = rpc.CONTEXT.copy()
        decoder = PYSONDecoder(ctx)
        action_ctx = decoder.decode(action.get('pyson_context') or '{}')
        domain = decoder.decode(action['pyson_domain'])
        screen = Screen(
            action['res_model'], mode=['tree'], view_ids=view_ids,
            domain=domain, context=action_ctx, readonly=True, limit=None,
            row_activate=self.menu_row_activate)
        # Use alternate view to not show search box
        screen.screen_container.alternate_view = True
        screen.switch_view(view_type=screen.current_view.view_type)

        self.menu.pack_start(
            screen.screen_container.alternate_viewport,
            expand=True, fill=True, padding=0)
        treeview = screen.current_view.treeview
        treeview.set_headers_visible(False)

        # Favorite column
        column = Gtk.TreeViewColumn()
        column.name = None
        column._type = None
        favorite_renderer = CellRendererClickablePixbuf()
        column.pack_start(favorite_renderer, expand=False)

        def favorite_setter(column, cell, store, iter_, user_data=None):
            menu = store.get_value(iter_, 0)
            favorite = menu.value.get('favorite')
            if favorite:
                icon = 'gnuhealth-star'
            elif favorite is False:
                icon = 'gnuhealth-star-border'
            else:
                icon = None
            if icon:
                pixbuf = common.IconFactory.get_pixbuf(
                    icon, Gtk.IconSize.MENU)
            else:
                pixbuf = None
            cell.set_property('pixbuf', pixbuf)
        column.set_cell_data_func(favorite_renderer, favorite_setter)

        def toggle_favorite(renderer, path, treeview):
            if treeview.props.window:
                self.toggle_favorite(renderer, path, treeview)
        favorite_renderer.connect(
            'clicked',
            lambda *a: GLib.idle_add(toggle_favorite, *a), treeview)
        # Unset fixed height mode to add column
        treeview.set_fixed_height_mode(False)
        treeview.set_property(
            'enable-grid-lines', Gtk.TreeViewGridLines.NONE)
        treeview.append_column(column)

        screen.search_filter()
        screen.display(set_cursor=True)
        self.menu_screen = screen

    def toggle_favorite(self, renderer, path, treeview):
        store = treeview.get_model()
        iter_ = store.get_iter(path)
        menu = store.get_value(iter_, 0)
        favorite = menu.value.get('favorite')
        if favorite:
            value = False
            method = 'unset'
        elif favorite is False:
            value = True
            method = 'set'
        else:
            return
        try:
            RPCExecute('model', self.menu_screen.model_name + '.favorite',
                       method, menu.id)
        except RPCException:
            return
        menu.value['favorite'] = value
        store.row_changed(path, iter_)
        self.favorite_unset()

    def win_set(self, page):
        current_page = self.notebook.get_current_page()
        page_num = self.notebook.page_num(page.widget)
        page.widget.props.visible = True
        self.notebook.set_current_page(page_num)
        # In order to focus the page
        if current_page == page_num:
            self._sig_page_changt(self.notebook, None, page_num)

    def win_add(self, page, hide_current=False):
        previous_page_id = self.notebook.get_current_page()
        previous_widget = self.notebook.get_nth_page(previous_page_id)
        if previous_widget and hide_current:
            page_id = previous_page_id + 1
        else:
            page_id = -1
        self.previous_pages[page] = previous_widget
        self.pages.append(page)
        hbox = Gtk.HBox(spacing=3)
        if page.icon:
            hbox.pack_start(
                common.IconFactory.get_image(
                    page.icon, Gtk.IconSize.SMALL_TOOLBAR),
                expand=False, fill=False, padding=0)
        name = page.name
        label = Gtk.Label(
            label=common.ellipsize(name.split(' / ')[-1], 20),
            halign=Gtk.Align.START)
        self.tooltips.set_tip(label, page.name)
        self.tooltips.enable()
        hbox.pack_start(label, expand=True, fill=True, padding=0)

        button = Gtk.Button()
        button.set_relief(Gtk.ReliefStyle.NONE)
        button.set_can_focus(False)
        button.add(common.IconFactory.get_image(
            'gnuhealth-close', Gtk.IconSize.MENU))
        self.tooltips.set_tip(button, _('Close Tab'))
        button.connect('clicked', self._sig_remove_book, page.widget)
        hbox.pack_start(button, expand=False, fill=False, padding=0)
        size = Gtk.IconSize.lookup(Gtk.IconSize.MENU)
        button.set_size_request(size.width, size.height)

        hbox.show_all()
        label_menu = Gtk.Label(
            label=page.name, halign=Gtk.Align.START)
        page.widget.props.margin = 3
        self.notebook.insert_page_menu(page.widget, hbox, label_menu, page_id)
        self.notebook.set_tab_reorderable(page.widget, True)
        self.notebook.set_current_page(page_id)

    def _sig_remove_book(self, widget, page_widget):
        for page in self.pages:
            if page.widget == page_widget:
                if not page.widget.props.sensitive:
                    return
                page_num = self.notebook.page_num(page.widget)
                self.notebook.set_current_page(page_num)
                res = page.sig_close()
                if not res:
                    return
        self._win_del(page_widget)

    def _win_del(self, page_widget=None):
        if page_widget:
            page_id = self.notebook.page_num(page_widget)
        else:
            page_id = int(self.notebook.get_current_page())
            page_widget = self.notebook.get_nth_page(page_id)
        if page_id != -1:
            page = None
            for i in range(len(self.pages)):
                if self.pages[i].widget == page_widget:
                    page = self.pages.pop(i)
                    break
            self.notebook.remove_page(page_id)

            next_page_id = -1
            to_pop = []
            for i in self.previous_pages:
                if self.previous_pages[i] == page_widget:
                    to_pop.append(i)
                if i.widget == page_widget:
                    if self.previous_pages[i]:
                        next_page_id = self.notebook.page_num(
                            self.previous_pages[i])
                    to_pop.append(i)
            to_pop.reverse()
            for i in to_pop:
                self.previous_pages.pop(i)

            if hasattr(page, 'destroy'):
                page.destroy()
            del page

            current_widget = self.notebook.get_nth_page(next_page_id)
            if current_widget:
                current_widget.props.visible = True
            self.notebook.set_current_page(next_page_id)
        if not self.pages and self.menu_screen:
            self.menu_screen.set_cursor()
        return self.notebook.get_current_page() != -1

    def get_page(self, page_id=None):
        if page_id is None:
            page_id = self.notebook.get_current_page()
        if page_id == -1:
            return None
        page_widget = self.notebook.get_nth_page(page_id)
        for page in self.pages:
            if page.widget == page_widget:
                return page
        return None

    def _sig_page_changt(self, notebook, page, page_num):
        self.last_page = self.current_page
        last_form = self.get_page(self.current_page)
        if last_form:
            for dialog in last_form.dialogs:
                dialog.hide()

        self.current_page = self.notebook.get_current_page()
        current_form = self.get_page(self.current_page)

        def set_cursor():
            if self.current_page == self.notebook.get_current_page():
                current_form.set_cursor()
        # Using idle_add because the Gtk.TreeView grabs the focus at the
        # end of the event
        GLib.idle_add(set_cursor)
        for dialog in current_form.dialogs:
            dialog.show()

    def _open_url(self, url):
        loads = partial(json.loads, object_hook=object_hook)
        urlp = urlparse(url)
        if not urlp.scheme == 'gnuhealth':
            return
        urlp = urlparse('http' + url[6:])
        database, path = list(
            map(unquote, (urlp.path[1:].split('/', 1) + [''])[:2]))
        if not path:
            return
        type_, path = (path.split('/', 1) + [''])[:2]
        params = {}
        if urlp.params:
            try:
                params.update(dict(parse_qsl(urlp.params,
                              strict_parsing=True)))
            except ValueError:
                return

        def open_model(path):
            attributes = {}
            model, path = (path.split('/', 1) + [''])[:2]
            if not model:
                return
            attributes = {}
            try:
                attributes['view_ids'] = loads(params.get('views', '[]'))
                if 'limit' in params:
                    attributes['limit'] = loads(params.get('limit', 'null'))
                attributes['name'] = loads(params.get('name', '""'))
                attributes['search_value'] = loads(
                    params.get('search_value', '[]'))
                attributes['domain'] = loads(params.get('domain', '[]'))
                attributes['context'] = loads(params.get('context', '{}'))
                attributes['context_model'] = params.get('context_model')
                attributes['tab_domain'] = loads(
                    params.get('tab_domain', '[]'))
            except ValueError:
                return
            if path:
                try:
                    attributes['res_id'] = int(path)
                except ValueError:
                    return
                attributes['mode'] = ['form', 'tree']
            try:
                Window.create(model, **attributes)
            except Exception:
                # Prevent crashing the client
                return

        def open_wizard(wizard):
            attributes = {}
            if not wizard:
                return
            try:
                attributes['data'] = loads(params.get('data', '{}'))
                attributes['direct_print'] = loads(
                    params.get('direct_print', 'false'))
                attributes['name'] = loads(params.get('name', '""'))
                attributes['window'] = loads(params.get('window', 'false'))
                attributes['context'] = loads(params.get('context', '{}'))
            except ValueError:
                return
            try:
                Window.create_wizard(wizard, **attributes)
            except Exception:
                # Prevent crashing the client
                return

        def open_report(report):
            attributes = {}
            if not report:
                return
            try:
                attributes['data'] = loads(params.get('data', '{}'))
                attributes['direct_print'] = loads(
                    params.get('direct_print', 'false'))
                attributes['context'] = loads(params.get('context', '{}'))
            except ValueError:
                return
            try:
                Action.exec_report(report, **attributes)
            except Exception:
                # Prevent crashing the client
                return

        def open_url():
            try:
                url = loads(params.get('url', 'false'))
            except ValueError:
                return
            if url:
                webbrowser.open(url, new=2)

        if type_ == 'model':
            open_model(path)
        elif type_ == 'wizard':
            open_wizard(path)
        elif type_ == 'report':
            open_report(path)
        elif type_ == 'url':
            open_url()

        self.window.present()

    def open_url(self, url):
        def idle_open_url():
            self._open_url(url)
            return False
        GLib.idle_add(idle_open_url)

    def show_notification(self, title, msg, priority=1):
        notification = Gio.Notification.new(title)
        notification.set_body(msg)
        notification.set_priority(_PRIORITIES[priority])
        if sys.platform != 'win32' or GLib.glib_version >= (2, 57, 0):
            self.send_notification(None, notification)

    # GNU Health client
    def get_main_title(self):
        user = CONFIG['login.login']
        host = CONFIG['login.host']
        db = CONFIG['login.db']
        conn_info = f'{user}@{host}/{db}'
        if CONFIG['client.title']:
            res = f"{CONFIG['client.title']} - {conn_info}"
        else:
            res = conn_info
        return res

    # GNUHEALTH block
    # Initialize GNU Health environment
    def init_gnuhealth_env(self):
        # Init GNU Health CLI
        self.cli = None
        self.cli = Gtk.Entry()
        # Maximum of 64 chars
        self.cli.set_max_length(64)
        self.cli.set_placeholder_text(_("GNU Health Command"))
        # CLI colors
        self.cli.modify_base(Gtk.StateType.NORMAL, Gdk.Color(6, 69, 72))
        self.cli.modify_text(Gtk.StateType.NORMAL, Gdk.Color(0, 0, 0))

        # Init the GNU Health Status Bar
        self.statusbar = None
        self.footer = None
        self.ghheader = None
        self.ghheader_contents = None
        self.footer_contents = None
        self.statusbar = Gtk.Statusbar()
        self.context_id = self.statusbar.get_context_id("Statusbar")
        self.ghheader = Gtk.HBox()
        self.ghheader.set_size_request(200, -1)

        # Main footer
        self.ghfooter = Gtk.HBox()

        # Create initally a grid for the footer
        self.ghfooter_contents = Gtk.Grid()

        # Create initally a grid for the header
        self.ghheader_contents = Gtk.Grid()

        # Set GNU Health Footer
        self.set_ghfooter()

        # Include the side menu toggle
        side_menu = Gtk.Button.new()
        side_menu.set_image(
            common.IconFactory.get_image('gnuhealth-menu',
                                         Gtk.IconSize.BUTTON))
        side_menu.connect('clicked', self.menu_toggle)

        # End of side-menu toggle
        # Favorite entries
        favorite = Gtk.MenuButton.new()
        favorite.set_relief(Gtk.ReliefStyle.NONE)
        favorite.set_image(common.IconFactory.get_image(
            'gnuhealth-bookmarks', Gtk.IconSize.BUTTON))
        self.menu_favorite = Gtk.Menu.new()
        favorite.set_popup(self.menu_favorite)
        favorite.connect('clicked', self.favorite_set)

        self.set_global_search()

        # CLI position at top
        self.ghheader_contents.add(side_menu)
        self.ghheader_contents.attach_next_to(self.cli, side_menu,
                                              Gtk.PositionType.RIGHT, 1, 1)
        self.ghheader_contents.attach_next_to(favorite, self.cli,
                                              Gtk.PositionType.RIGHT, 1, 1)

        self.ghheader_contents.attach_next_to(
            self.global_search_entry, favorite, Gtk.PositionType.RIGHT, 1, 1)

        self.ghheader.pack_start(self.ghheader_contents, expand=True,
                                 fill=True, padding=10)
        self.vbox.pack_start(self.ghheader,
                             expand=False, fill=False, padding=10)

        self.vbox.reorder_child(self.ghheader, 1)

        self.ghheader.show_all()

    # Add GNU Health Command Line and Status Bar in footer
    def set_ghfooter(self):
        cli = self.cli
        statusbar = self.statusbar
        ghfooter_contents = self.ghfooter_contents

        ghfooter_contents.add(statusbar)

        self.ghfooter.pack_start(ghfooter_contents, expand=True, fill=True,
                                 padding=0)

        # Pack the footer table in main vbox
        self.vbox.pack_start(self.ghfooter, expand=False,
                             fill=False, padding=0)

        self.ghfooter.show_all()
        self.vbox.show_all()

        cli.connect('activate', self.activate_cli, self.context_id)

    # Parse the CLI arguments
    def activate_cli(self, widget, data):
        # Make the command case-insensitive
        command = widget.get_text().lower()
        if (command == 'sysinfo'):
            info = f"Request : {command} \n"
            client_info = '\n********* CLIENT INFORMATION *********\n\n'
            server_info = '\n********* SERVER INFORMATION *********\n\n'
            client_info = f"{client_info} {self.get_host_info()}\n"

            server_info = server_info + \
                RPCExecute('model', 'gnuhealth.command', command)

            info = info + client_info + server_info
            self.activity_log_entry(info, "info")
        else:
            if command:
                self.gnuhealth_cmd(command)

    # GNU Health system activity and status logger
    def activity_log_entry(self, msg, msg_type):
        time_stamp = time.strftime("%Y/%m/%d - %H:%M:%S")
        if (msg_type == "info"):
            mtype = "[INFO]"
        elif (msg_type == "info"):
            mtype = "[WARN]"
        else:
            mtype = "[ERROR]"

        log_entry = time_stamp + "  " + mtype + ": " + msg + "\n"

        next_line = Activity.textbuffer.get_end_iter()

        Activity.textbuffer.insert(next_line, log_entry)

    # Reset and clean GNU Health environment
    def reset_gnuhealth_env(self):
        if self.cli:
            self.cli.destroy()

        if self.statusbar:
            self.statusbar.remove_all(self.context_id)

        if self.footer_contents:
            self.footer_contents.destroy()

        if self.footer:
            self.footer.destroy()

    def get_host_info(self):
        info = ""
        os_header = "\n-- Operating System / Distribution --\n"
        uname = platform.uname()
        client_version = f"GNU Health GTK Client version: {__version__}"
        pversion = f"Python version : {platform.python_version()}"
        with open('/etc/os-release') as f:
            os_version = f.readlines()
        # Get OS version and related info.
        os_info = ''
        with open('/etc/os-release') as f:
            os_release = f.readlines()
            for line in os_release:
                os_info = os_info + str(line)
        # Get relevant environment variables

        info = f"{info} {pversion}\n {client_version}\n" \
            f"{os_header} {os_version} {os_info}\n" \
            f"Platform / Kernel Info: {str(uname)}"

        return info

    def gnuhealth_cmd(self, command):

        cmd = command.split()[0]
        args = command.split()[1:]
        domain_name = None
        search_string = '%'

        res = RPCExecute('model', 'gnuhealth.command', 'search_read',
                         [('name', '=', cmd)], 0, 1, None,
                         ['name', 'model', 'label', 'domain'])

        if (res):
            res = res[0]
            if (res['domain']):
                domain_name = ast.literal_eval(res['domain'])
            model_name = res['model']
            label = res['label']
            if args:
                arg1 = args.pop()
                search_string = "%" + arg1 + "%"

            Window.create(
                model_name,
                mode=['tree', 'form'],
                name=label,
                domain=domain_name,
                limit=CONFIG['client.limit'],
                search_value=[('rec_name', 'ilike', search_string)]
            )

        else:
            common.message(_('Command not found.'))

    # Toggle GNU Health Activity window
    def activity_toggle(self):
        if Activity().activity_window.get_visible():
            Activity().activity_window.hide()
        else:
            Activity().activity_window.show_all()



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import gettext
import webbrowser

import gnuhealth.rpc as rpc
from gnuhealth.common import (
    RPCException, RPCExecute, file_open, file_write, message, selection)
from gnuhealth.pyson import PYSONDecoder

_ = gettext.gettext


class Action(object):

    @staticmethod
    def exec_report(name, data, direct_print=False, context=None):
        if context is None:
            context = {}
        else:
            context = context.copy()
        context['direct_print'] = direct_print
        ids = data.get('ids', [])

        def callback(result):
            try:
                result = result()
            except RPCException:
                return
            type, data, print_p, name = result
            if not print_p and direct_print:
                print_p = True
            fp_name = file_write((name, type), data)
            file_open(fp_name, type, print_p=print_p)
        RPCExecute(
            'report', name, 'execute', ids, data, context=context,
            callback=callback)

    @staticmethod
    def execute(action, data, context=None, keyword=False):
        if isinstance(action, int):
            # Must be executed synchronously to avoid double execution
            # on double click.
            action = RPCExecute(
                'model', 'ir.action', 'get_action_value', action,
                context=context)
        if keyword:
            keywords = {
                'ir.action.report': 'form_report',
                'ir.action.wizard': 'form_action',
                'ir.action.act_window': 'form_relate',
            }
            action.setdefault('keyword', keywords.get(action['type'], ''))
        Action._exec_action(action, data, context=context)

    @staticmethod
    def _exec_action(action, data=None, context=None):
        from gnuhealth.gui.window import Window
        if context is None:
            context = {}
        else:
            context = context.copy()
        if data is None:
            data = {}
        else:
            data = data.copy()
        if 'type' not in (action or {}):
            return

        context.pop('active_id', None)
        context.pop('active_ids', None)
        context.pop('active_model', None)

        def add_name_suffix(name, context=None):
            if not data.get('ids') or not data.get('model'):
                return name
            max_records = 5
            ids = list(filter(lambda id: id >= 0, data['ids']))[:max_records]
            if not ids:
                return name
            rec_names = RPCExecute('model', data['model'],
                                   'read', ids, ['rec_name'],
                                   context=context)
            name_suffix = _(', ').join([x['rec_name'] for x in rec_names])
            if len(data['ids']) > len(ids):
                name_suffix += _(',...')
            if name_suffix:
                return _('%s (%s)') % (name, name_suffix)
            else:
                return name

        data['action_id'] = action['id']
        params = {
            'icon': action.get('icon.rec_name') or '',
        }
        if action['type'] == 'ir.action.act_window':
            if action.get('views', []):
                params['view_ids'] = [x[0] for x in action['views']]
                params['mode'] = [x[1] for x in action['views']]
            elif action.get('view_id', False):
                params['view_ids'] = [action['view_id'][0]]

            action.setdefault('pyson_domain', '[]')
            ctx = {
                'active_model': data.get('model'),
                'active_id': data.get('id'),
                'active_ids': data.get('ids') or [],
            }
            ctx.update(rpc.CONTEXT)
            ctx['_user'] = rpc._USER
            decoder = PYSONDecoder(ctx)
            params['context'] = context.copy()
            params['context'].update(
                decoder.decode(action.get('pyson_context') or '{}'))
            ctx.update(params['context'])

            ctx['context'] = ctx
            decoder = PYSONDecoder(ctx)
            params['domain'] = decoder.decode(action['pyson_domain'])
            params['order'] = decoder.decode(action['pyson_order'])
            params['search_value'] = decoder.decode(
                action['pyson_search_value'] or '[]')
            params['tab_domain'] = [
                (n, decoder.decode(d), c) for n, d, c in action['domains']]

            name = action.get('name', '')
            if action.get('keyword', ''):
                name = add_name_suffix(name, params['context'])
            params['name'] = name

            res_model = action.get('res_model', data.get('res_model'))
            params['res_id'] = action.get('res_id', data.get('res_id'))
            params['context_model'] = action.get('context_model')
            params['context_domain'] = action.get('context_domain')
            limit = action.get('limit')
            if limit is not None:
                params['limit'] = limit

            Window.create(res_model, **params)
        elif action['type'] == 'ir.action.wizard':
            params['context'] = context
            params['window'] = action.get('window')
            name = action.get('name', '')
            if action.get('keyword', 'form_action') == 'form_action':
                name = add_name_suffix(name, context)
            params['name'] = name
            Window.create_wizard(action['wiz_name'], data, **params)
        elif action['type'] == 'ir.action.report':
            params['direct_print'] = action.get('direct_print', False)
            params['context'] = context
            del params['icon']
            Action.exec_report(action['report_name'], data, **params)

        elif action['type'] == 'ir.action.url':
            if action['url']:
                webbrowser.open(action['url'], new=2)

    @staticmethod
    def exec_keyword(keyword, data=None, context=None, warning=True,
                     alwaysask=False):
        actions = []
        model_id = data.get('id', False)
        try:
            actions = RPCExecute('model', 'ir.action.keyword',
                                 'get_keyword',
                                 keyword, (data['model'], model_id))
        except RPCException:
            return False

        keyact = {}
        for action in actions:
            keyact[action['name'].split(' / ')[-1]] = action

        res = selection(_('Select your action'), keyact, alwaysask=alwaysask)
        if res:
            (name, action) = res
            Action.execute(action, data, context=context)
            return (name, action)
        elif not len(keyact) and warning:
            message(_('No action defined.'))
        return False



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from unittest import TestCase

from gnuhealth.common.selection import freeze_value


class SelectionTestCase(TestCase):
    "Test common selection"

    def test_freeze_value(self):
        "Test freeze_value"
        self.assertEqual(freeze_value({'foo': 'bar'}), (('foo', 'bar'),))
        self.assertEqual(freeze_value([1, 42, 2, 3]), (1, 42, 2, 3))
        self.assertEqual(freeze_value('foo'), 'foo')
        self.assertEqual(
            freeze_value({'foo': {'bar': 42}}), (('foo', (('bar', 42),)),))



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import gettext
import logging

from gi.repository import Gdk, Gtk, Pango

import gnuhealth.common as common
from gnuhealth.common import (
    GNUHEALTH_ICON, RPCContextReload, RPCException, RPCExecute)
from gnuhealth.common.button import Button
from gnuhealth.common.widget_style import widget_class
from gnuhealth.exceptions import GNUHealthServerError
from gnuhealth.gui import Main
from gnuhealth.gui.window.nomodal import NoModal
from gnuhealth.gui.window.view_form.screen import Screen

from .infobar import InfoBar
from .tabcontent import TabContent

_ = gettext.gettext
logger = logging.getLogger(__name__)


class Wizard(InfoBar):

    def __init__(self, name=''):
        super(Wizard, self).__init__()
        self.widget = Gtk.VBox(spacing=3)
        self.widget.show()
        self.name = name or _('Wizard')
        self.id = None
        self.ids = None
        self.action = None
        self.action_id = None
        self.direct_print = False
        self.email_print = False
        self.email = False
        self.context = None
        self.states = {}
        self.response2button = {}
        self.__processing = False
        self.__waiting_response = False
        self.session_id = None
        self.start_state = None
        self.end_state = None
        self.screen = None
        self.screen_state = None
        self.state = None

    def run(self, action, data, direct_print=False, email_print=False,
            email=None, context=None):
        self.action = action
        self.action_id = data.get('action_id')
        self.id = data.get('id')
        self.ids = data.get('ids')
        self.model = data.get('model')
        self.direct_print = direct_print
        self.email_print = email_print
        self.email = email
        self.context = context.copy() if context is not None else {}
        self.context['active_id'] = self.id
        self.context['active_ids'] = self.ids
        self.context['active_model'] = self.model
        self.context['action_id'] = self.action_id

        def callback(result):
            try:
                result = result()
            except RPCException:
                self.destroy()
                return
            self.session_id, self.start_state, self.end_state = result
            self.state = self.start_state
            self.process()
        RPCExecute('wizard', action, 'create', callback=callback)

    def process(self):
        from gnuhealth.action import Action
        if self.__processing or self.__waiting_response:
            return
        self.__processing = True

        ctx = self.context.copy()
        if self.screen:
            data = {
                self.screen_state: self.screen.get_on_change_value(),
            }
        else:
            data = {}

        def callback(result):
            try:
                result = result()
            except RPCException as rpc_exception:
                if (not isinstance(rpc_exception.exception,
                                   GNUHealthServerError)
                        or not self.screen):
                    self.state = self.end_state
                    self.end()
                self.__processing = False
                return

            if 'view' in result:
                self.clean()
                view = result['view']
                self.update(view['fields_view'], view['buttons'])

                self.screen.new(default=False)
                if 'defaults' in view:
                    self.screen.current_record.set_default(view['defaults'])
                if 'values' in view:
                    self.screen.current_record.set(view['values'])
                self.update_buttons()
                self.screen.set_cursor()

                self.screen_state = view['state']
                self.__waiting_response = True
            else:
                self.state = self.end_state

            def execute_actions():
                for action in result.get('actions', []):
                    for k, v in [
                            ('direct_print', self.direct_print),
                            ('email_print', self.email_print),
                            ('email', self.email),
                    ]:
                        action[0].setdefault(k, v)
                    context = self.context.copy()
                    # Remove wizard keys added by run
                    del context['active_id']
                    del context['active_ids']
                    del context['active_model']
                    del context['action_id']
                    Action.execute(*action, context=context)

            if self.state == self.end_state:
                self.end(lambda *a: execute_actions())
            else:
                execute_actions()
            self.__processing = False

        RPCExecute('wizard', self.action, 'execute', self.session_id, data,
                   self.state, context=ctx, callback=callback)

    def destroy(self, action=None):
        if self.screen:
            self.screen.destroy()

    def end(self, callback=None):
        def end_callback(action):
            self.destroy(action=action())
            if callback:
                callback()
        try:
            RPCExecute('wizard', self.action, 'delete', self.session_id,
                       process_exception=False, callback=end_callback)
        except Exception:
            logger.warn(
                "Unable to delete session %s of wizard %s",
                self.session_id, self.action, exc_info=True)

    def clean(self):
        for widget in self.widget.get_children():
            self.widget.remove(widget)
        self.states = {}

    def response(self, widget, response):
        self.__waiting_response = False
        button_attrs = self.response2button[response].attrs
        state = button_attrs.get('state', self.end_state)
        self.screen.current_view.set_value()
        if (button_attrs.get('validate', True)
                and not self.screen.current_record.validate()):
            self.screen.display(set_cursor=True)
            self.info_bar_add(
                self.screen.invalid_message(), Gtk.MessageType.ERROR)
            return
        self.info_bar_clear()
        self.state = state
        self.process()

    def _get_button(self, definition):
        button = Button(definition)
        self.states[definition['state']] = button
        response = len(self.states)
        self.response2button[response] = button
        button.show()
        return button

    def record_modified(self):
        self.update_buttons()
        self.info_bar_refresh()

    def update_buttons(self):
        record = self.screen.current_record
        for button in self.states.values():
            button.state_set(record)

    def update(self, view, buttons):
        tooltips = common.Tooltips()
        for button in buttons:
            self._get_button(button)

        if self.screen:
            self.screen.windows.remove(self)

        self.screen = Screen(view['model'], mode=[], context=self.context)
        self.screen.add_view(view)
        self.screen.switch_view()
        self.screen.widget.show()
        self.screen.windows.append(self)

        title = Gtk.Label(
            label=common.ellipsize(self.name, 80),
            halign=Gtk.Align.START, margin=5,
            ellipsize=Pango.EllipsizeMode.END)
        tooltips.set_tip(title, self.name)
        title.set_size_request(0, -1)  # Allow overflow
        title.show()

        hbox = Gtk.HBox()
        hbox.pack_start(title, expand=True, fill=True, padding=0)
        hbox.show()

        frame = Gtk.Frame()
        frame.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        widget_class(frame, 'wizard-title', True)
        frame.add(hbox)
        frame.show()

        self.widget.pack_start(frame, expand=False, fill=True, padding=3)

        viewport = Gtk.Viewport()
        viewport.set_shadow_type(Gtk.ShadowType.NONE)
        viewport.add(self.screen.widget)
        viewport.show()
        self.scrolledwindow = Gtk.ScrolledWindow()
        self.scrolledwindow.set_shadow_type(Gtk.ShadowType.NONE)
        self.scrolledwindow.set_policy(
            Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.scrolledwindow.add(viewport)
        self.scrolledwindow.show()

        self.widget.pack_start(
            self.scrolledwindow, expand=True, fill=True, padding=0)

        self.widget.pack_start(
            self.create_info_bar(), expand=False, fill=True, padding=0)


class WizardForm(Wizard, TabContent):
    "Wizard"

    def __init__(self, name=''):
        super(WizardForm, self).__init__(name=name)
        self.hbuttonbox = Gtk.HButtonBox()
        self.hbuttonbox.set_spacing(5)
        self.hbuttonbox.set_layout(Gtk.ButtonBoxStyle.END)
        self.hbuttonbox.show()
        self.dialogs = []

        self.handlers = {
            'but_close': self.sig_close
        }

    def clean(self):
        super(WizardForm, self).clean()
        for button in self.hbuttonbox.get_children():
            self.hbuttonbox.remove(button)

    def _get_button(self, state):
        button = super(WizardForm, self)._get_button(state)
        response = len(self.states)
        button.connect('clicked', self.response, response)
        self.hbuttonbox.pack_start(button, expand=True, fill=True, padding=0)
        return button

    def update(self, view, buttons):
        super(WizardForm, self).update(view, buttons)
        self.widget.pack_start(
            self.hbuttonbox, expand=False, fill=True, padding=0)

    def sig_close(self):
        if self.end_state in self.states:
            self.states[self.end_state].clicked()
        return self.state == self.end_state

    def destroy(self, action=None):
        super(WizardForm, self).destroy(action=action)
        if action == 'reload menu':
            RPCContextReload(Main().sig_win_menu)
        elif action == 'reload context':
            RPCContextReload()

    def end(self, callback=None):
        super(WizardForm, self).end(callback=callback)
        Main()._win_del(self.widget)

    def set_cursor(self):
        if self.screen:
            self.screen.set_cursor(reset_view=False)


class WizardDialog(Wizard, NoModal):

    def __init__(self, name=''):
        Wizard.__init__(self, name=name)
        NoModal.__init__(self)
        self.dia = Gtk.Dialog(
            title=self.name, transient_for=self.parent,
            destroy_with_parent=True)
        Main().add_window(self.dia)
        self.dia.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
        self.dia.set_icon(GNUHEALTH_ICON)
        self.dia.set_deletable(False)
        self.dia.connect('delete-event', lambda *a: True)
        self.dia.connect('close', self.close)
        self.dia.connect('response', self.response)

        self.accel_group = Gtk.AccelGroup()
        self.dia.add_accel_group(self.accel_group)

        self._buttons = set()

        self.dia.vbox.pack_start(
            self.widget, expand=True, fill=True, padding=0)

        self.register()

    def clean(self):
        super(WizardDialog, self).clean()
        while self._buttons:
            button = self._buttons.pop()
            button.get_parent().remove(button)

    def _get_button(self, definition):
        button = super(WizardDialog, self)._get_button(definition)
        response = len(self.states)
        self.dia.add_action_widget(button, response)
        self._buttons.add(button)
        if definition['default']:
            button.add_accelerator(
                'clicked', self.accel_group, Gdk.KEY_Return,
                Gdk.ModifierType.CONTROL_MASK, Gtk.AccelFlags.VISIBLE)
            button.get_style_context().add_class(
                Gtk.STYLE_CLASS_SUGGESTED_ACTION)
            button.set_can_default(True)
            button.grab_default()
            self.dia.set_default_response(response)
        return button

    def update(self, view, buttons):
        super(WizardDialog, self).update(view, buttons)
        current_view = self.screen.current_view
        self.scrolledwindow.set_policy(
            Gtk.PolicyType.NEVER, Gtk.PolicyType.NEVER)
        if current_view.scroll:
            current_view.scroll.set_policy(
                Gtk.PolicyType.NEVER, Gtk.PolicyType.NEVER)
        self.show()
        self.scrolledwindow.set_policy(
            Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        if current_view.scroll:
            current_view.scroll.set_policy(
                Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)

    def destroy(self, action=None):
        super(WizardDialog, self).destroy(action=action)
        self.dia.destroy()
        NoModal.destroy(self)
        main = Main()
        if self.parent == main.window:
            current_form = main.get_page()
            if current_form:
                for dialog in current_form.dialogs:
                    dialog.show()
        if self.page.dialogs:
            dialog = self.page.dialogs[-1]
        else:
            dialog = self.page
        screen = getattr(dialog, 'screen', None)
        if self.sensible_widget == main.window:
            screen = main.menu_screen
        if screen:
            if (screen.current_record
                    and self.sensible_widget != main.window):
                if screen.model_name == self.model:
                    ids = self.ids
                else:
                    # Wizard run from a children record so reload parent record
                    ids = [screen.current_record.id]
                screen.reload(ids, written=True)
            if action:
                screen.client_action(action)

    def close(self, widget, event=None):
        widget.stop_emission_by_name('close')
        if self.end_state in self.states:
            self.states[self.end_state].clicked()
        return True

    def show(self):
        if not self.screen:
            return
        view = self.screen.current_view
        if view.view_type == 'form':
            expand = False
            for name in view.get_fields():
                for widget in view.widgets[name]:
                    if widget.expand:
                        expand = True
                        break
                if expand:
                    break
        else:
            expand = True
        if expand:
            width, height = self.default_size()
        else:
            width, height = -1, -1
        self.dia.set_default_size(width, height)
        # reshow with initial size
        self.dia.hide()
        self.dia.unrealize()
        self.dia.show()
        width, height = self.dia.get_size()
        screen = self.dia.get_screen()
        self.dia.resize(
            min(width, screen.width()), min(height, screen.height()))

    def hide(self):
        self.dia.hide()



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from .window import Window

__all__ = [Window]



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from gi.repository import GObject, Gtk


class CellRendererClickablePixbuf(Gtk.CellRendererPixbuf):
    __gsignals__ = {
        'clicked': (GObject.SignalFlags.RUN_LAST, GObject.TYPE_NONE,
                    (GObject.TYPE_STRING, )),
    }

    def __init__(self):
        Gtk.CellRendererPixbuf.__init__(self)
        self.set_property('mode', Gtk.CellRendererMode.ACTIVATABLE)

    def do_activate(
            self, event, widget, path, background_area, cell_area, flags):
        self.emit('clicked', path)


GObject.type_register(CellRendererClickablePixbuf)



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import gettext

from gi.repository import Gtk

from gnuhealth.common import IconFactory

_ = gettext.gettext


class Button(Gtk.Button):

    def __init__(self, attrs=None):
        self.attrs = attrs or {}
        self.label = '_' + attrs.get('string', '').replace('_', '__')
        super(Button, self).__init__(label=self.label, stock=None,
                                     use_underline=True)
        self._set_icon(attrs.get('icon'))

    def _set_icon(self, stock):
        self.set_always_show_image(bool(stock))
        image = self.get_image()
        if not image and not stock:
            return
        if not stock:
            self.set_image(None)
            return
        self.set_image(IconFactory.get_image(stock, Gtk.IconSize.BUTTON))

    def state_set(self, record):
        if record:
            states = record.expr_eval(self.attrs.get('states', {}))
        else:
            states = {}
        if states.get('invisible', False):
            self.hide()
        else:
            self.show()
        self.set_sensitive(not states.get('readonly', False))
        self._set_icon(states.get('icon', self.attrs.get('icon')))

        if self.attrs.get('rule'):
            label = self.label
            tip = self.attrs.get('help', '')
            if record:
                clicks = record.get_button_clicks(self.attrs['name'])
                if clicks:
                    label += ' (%s)' % len(clicks)
                    if tip:
                        tip += '\n'
                    tip += _('By: ') + _(', ').join(iter(clicks.values()))
            self.set_label(label)
            self.set_tooltip_text(tip)

        if self.attrs.get('type', 'class') == 'class':
            parent = record.parent if record else None
            while parent:
                if parent.modified:
                    self.set_sensitive(False)
                    break
                parent = parent.parent



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from weakref import WeakKeyDictionary

from gi.repository import Gtk

from gnuhealth.common.htmltextbuffer import normalize_markup
from gnuhealth.common.richtext import (
    add_toolbar, get_content, register_format, set_content)
from gnuhealth.config import CONFIG

from .textbox import TextBox


class RichTextBox(TextBox):

    def __init__(self, view, attrs):
        super(RichTextBox, self).__init__(view, attrs)
        self.toolbar = None
        self.tag_widgets = WeakKeyDictionary()
        self.tags = {}
        self.colors = {}
        if int(self.attrs.get('toolbar', 1)):
            self.toolbar = add_toolbar(self.textview)
            self.toolbar.set_style({
                'default': False,
                'both': Gtk.ToolbarStyle.BOTH,
                'text': Gtk.ToolbarStyle.TEXT,
                'icons': Gtk.ToolbarStyle.ICONS,
            }[CONFIG['client.toolbar']])
            self.widget.pack_start(
                self.toolbar, expand=False, fill=True, padding=0)

    def _get_textview(self):
        textview = super(RichTextBox, self)._get_textview()
        register_format(textview)
        return textview

    def translate_widget(self):
        widget = super(RichTextBox, self).translate_widget()
        textview = widget.get_children()[-1].get_child()
        if self.toolbar:
            widget.pack_start(
                add_toolbar(textview), expand=False, fill=True, padding=0)
        return widget

    def translate_widget_set_readonly(self, widget, value):
        super(RichTextBox, self).translate_widget_set_readonly(widget, value)
        if self.toolbar:
            toolbar = widget.get_children()[0]
            for n in range(toolbar.get_n_items()):
                tool = toolbar.get_nth_item(n)
                tool.set_sensitive(not value)

    def set_value(self):
        # avoid modification of not normalized value
        value = self.get_value()
        prev_value = self.field.get_client(self.record) or ''
        if value == normalize_markup(prev_value):
            value = prev_value
        self.field.set_client(self.record, value)

    @property
    def modified(self):
        if self.record and self.field:
            value = normalize_markup(self.field.get_client(self.record) or '')
            return value != self.get_value()
        return False

    def set_buffer(self, value, textview):
        set_content(textview, value)

    def get_buffer(self, textview):
        return get_content(textview)

    def _readonly_set(self, value):
        super(RichTextBox, self)._readonly_set(value)
        if self.toolbar:
            self.toolbar.set_sensitive(not value)
